import requests

from ximilar.client import RestClient
from ximilar.client.constants import *

READ_ENDPOINT = "ocr/v2/read"
READ_GPT_ENDPOINT = "ocr/v2/read_gpt"


class OCRClient(RestClient):
    def construct_data(self, records, **kwargs):
        if len(records) == 0:
            raise Exception("Please specify at least on record when using ocr endpoint.")

        data = {RECORDS: self.preprocess_records(records)}

        if kwargs:
            data.update(kwargs)

        return data

    def read(self, records, lang="en", **kwargs):
        """
        Call the ocr endpoint.
        :param records: list of json records
        :param lang: specify the language of the text that is probably on image, defaults to "en"
        :return: json result data from the API
        """
        data = self.construct_data(records, lang=lang, **kwargs)
        result = self.post(READ_ENDPOINT, data=data)
        return result

    def read_gpt(self, records, lang="en", **kwargs):
        """
        Call the ocr + gpt endpoint.
        :param records: list of json records
        :param lang: specify the language of the text that is probably on image, defaults to "en"
        :return: json result data from the API
        """
        data = self.construct_data(records, lang=lang, **kwargs)
        result = self.post(READ_GPT_ENDPOINT, data=data)
        return result


if __name__ == "__main__":
    import sys

    client = OCRClient(sys.argv[1])
    print(client.read([{"_url": "https://images.ximilar.com/examples/cards/mew_pokemon.jpeg"}]))
    print(
        client.read_gpt(
            [
                {
                    "_url": "https://images.ximilar.com/examples/cards/mew_pokemon.jpeg",
                    "prompt": "based on the following result from ocr system what is the name and type of the card as json result ({'name':'', 'type': ''})",
                }
            ]
        )
    )



import requests
import json

from ximilar.client import RecognitionClient
from ximilar.client.constants import *
from ximilar.client.exceptions import XimilarClientInvalidDataException


REQUEST_ENDPOINT = "account/v2/request/"


class AsyncRClient(RecognitionClient):
    def __init__(
        self,
        token,
        endpoint=ENDPOINT,
        workspace=DEFAULT_WORKSPACE,
    ):
        super().__init__(
            token=token,
            endpoint=endpoint,
            workspace=workspace,
            resource_name=None,
        )

    def get_requests(self, page_url=None, type=None, status=None):
        url = page_url if page_url else REQUEST_ENDPOINT + "?page=1"

        if type is not None:
            url += "&type=" + str(type)

        if status is not None:
            url += "&status=" + str(status)

        result = self.get(url)
        return (
            [AsynchronousRequest(self.token, self.endpoint, self.workspace, ar_json) for ar_json in result[RESULTS]],
            result[NEXT],
            {"count": result["count"], STATUS: "ok"},
        )

    def get_request(self, id):
        result = self.get(REQUEST_ENDPOINT + id)
        return AsynchronousRequest(self.token, self.endpoint, self.workspace, result), STATUS_OK

    def submit(self, request, type, **kwargs):
        data = request
        data["type"] = type
        data.update(kwargs)
        result = self.post(REQUEST_ENDPOINT, data=data)
        return AsynchronousRequest(self.token, self.endpoint, self.workspace, result)


class AsynchronousRequest(AsyncRClient):
    def __init__(self, token, endpoint, workspace, async_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace)

        self.id = async_json["id"]
        self.status = async_json["status"]
        self.type = async_json["type"]
        self.response = async_json["response"] if "response" in async_json else None

    def __str__(self) -> str:
        return self.id

    def update(self):
        if self.response is None:
            if self.check_status() == "DONE":
                request, _ = self.get_request(self.id)
                self.status = request.status
                self.response = request.response
                return True
        return False

    def check_status(self):
        result = self.get(REQUEST_ENDPOINT + self.id + "/status")
        self.status = result["status"]
        return self.status



from ximilar.client import RecognitionClient
from ximilar.client.recognition import Image, IMAGE_ENDPOINT
from ximilar.client.constants import *

from urllib.request import urlopen


OBJECT_ENDPOINT = "detection/v2/object/"
LABEL_ENDPOINT = "detection/v2/label/"
TASK_ENDPOINT = "detection/v2/task/"
DETECT_ENDPOINT = "detection/v2/detect/"


class DetectionClient(RecognitionClient):
    def __init__(
        self,
        token,
        endpoint=ENDPOINT,
        workspace=DEFAULT_WORKSPACE,
        max_image_size=1024,
        resource_name=CUSTOM_OBJECT_DETECTION,
    ):
        super().__init__(
            token=token,
            endpoint=endpoint,
            workspace=workspace,
            max_image_size=max_image_size,
            resource_name=resource_name,
        )
        self.PREDICT_ENDPOINT = DETECT_ENDPOINT

    def get_object(self, object_id):
        """
        Getting Bounding Box/Detection Object by id.
        :param object_id: uuid
        :return: object, status
        """
        object_json = self.get(OBJECT_ENDPOINT + object_id)
        if ID not in object_json:
            status = {STATUS: object_json[DETAIL]} if DETAIL in object_json else {STATUS: "Not Found"}
            return None, status
        return DetectionObject(self.token, self.endpoint, object_json), RESULT_OK

    def get_label(self, label_id):
        """
        Getting Bounding Box/Detection Object by id.
        :param object_id: uuid
        :return: object, status
        """
        label_json = self.get(LABEL_ENDPOINT + label_id)
        if ID not in label_json:
            return None, {STATUS: "label with id not found"}
        return DetectionLabel(self.token, self.endpoint, label_json), RESULT_OK

    def get_task(self, task_id):
        """
        Getting Detection Task by id.
        :param object_id: uuid
        :return: task, status
        """
        task_json = self.get(TASK_ENDPOINT + task_id)
        if ID not in task_json:
            status = {STATUS: task_json[DETAIL]} if DETAIL in task_json else {STATUS: "Not Found"}
            return None, status
        return DetectionTask(self.token, self.endpoint, task_json, self.max_image_size), RESULT_OK

    def get_model(self, model_id):
        pass

    def remove_object(self, object_id):
        """
        Removes detection object by id/uuid.
        """
        return self.delete(OBJECT_ENDPOINT + object_id)

    def remove_label(self, label_id):
        """
        Removes detection label by id/uuid.
        """
        return self.delete(LABEL_ENDPOINT + label_id)

    def remove_task(self, task_id):
        """
        Removes detection task by id/uuid.
        """
        return self.delete(TASK_ENDPOINT + task_id)

    def remove_model(self, object_id):
        pass

    def get_objects(self, page_url=None, object_label=None, detection_label=None):
        """
        Get paginated result of all Detection Objects in your workspace.
        :param page_url: optional, select the specific page of images, default first page
        :param object_label: recognition label that must be present in objects
        :return: (list of images, next_page)
        """
        url = (
            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
            if page_url
            else OBJECT_ENDPOINT + "?page=1"
        )
        if object_label:
            url += "&object_labels=" + object_label
        if detection_label:
            url += "&object_detection_label=" + detection_label

        result = self.get(url)
        return (
            [DetectionObject(self.token, self.endpoint, object_json) for object_json in result[RESULTS]],
            result[NEXT],
            RESULT_OK,
        )

    def get_objects_of_image(self, image_id):
        """
        Get all Detection Objects which are located on image.
        :param image_id: uuid of image
        :return: list, result
        """
        objects, status = self.get_all_paginated_items(OBJECT_ENDPOINT + "?image=" + image_id)

        if not objects and status[STATUS] == STATUS_ERROR:
            return None, status

        return [DetectionObject(self.token, self.endpoint, o_json) for o_json in objects], RESULT_OK

    def get_all_tasks(self, suffix=""):
        """
        Get all Detection Tasks of the user(user is specified by client key).
        :return: List of Tasks
        """
        tasks, status = self.get_all_paginated_items(TASK_ENDPOINT + suffix)

        if not tasks and status[STATUS] == STATUS_ERROR:
            return None, status

        return [DetectionTask(self.token, self.endpoint, t_json, self.max_image_size) for t_json in tasks], RESULT_OK

    def get_tasks_by_name(self, name):
        """
        Get all tasks with the name.
        """
        tasks, result = self.get_all_tasks()

        tasks_to_return = []
        if result[STATUS] == STATUS_OK:
            for task in tasks:
                if task.name == name:
                    tasks_to_return.append(task)
        else:
            return None, result

        if len(tasks_to_return) > 0:
            return tasks_to_return, result

        return None, {STATUS: "Task with this name not found!"}

    def get_all_labels(self, suffix=""):
        """
        Get all Detection Labels of the user(user is specified by client key).
        :return: List of labels
        """
        labels, status = self.get_all_paginated_items(LABEL_ENDPOINT + suffix)

        if not labels and status[STATUS] == STATUS_ERROR:
            return None, status

        return [DetectionLabel(self.token, self.endpoint, l_json) for l_json in labels], RESULT_OK

    def get_label_by_name(self, name):
        """
        Get label with specified name which also belongs to this task.
        """
        labels, result = self.get_all_labels()
        if result[STATUS] == STATUS_OK:
            for label in labels:
                if label.name == name:
                    return label, RESULT_OK
        else:
            return None, result

        return None, {STATUS: "Label with this name not found!"}

    def create_task(self, name, description=None):
        """
        Create detection task with given name.
        :param name: name of the task
        :param description: description of the label
        :return: Task object, status
        """
        task_json = self.post(TASK_ENDPOINT, data={NAME: name, DESCRIPTION: description})
        if ID not in task_json:
            msg = task_json[DETAIL] if DETAIL in task_json else "unexpected error"
            return None, {STATUS: msg}
        return DetectionTask(self.token, self.endpoint, task_json, self.max_image_size), RESULT_OK

    def create_label(self, name, description=None, color="#FFFFFF", output_name=None):
        """
        Create detection label with given name.
        :param name: name of the label
        :param description: description of the label
        :param color: color (hexadecimal color code) of the label
        :return: Label object, status
        """
        label_json = self.post(
            LABEL_ENDPOINT, data={NAME: name, DESCRIPTION: description, COLOR: color, OUTPUT_NAME: output_name}
        )
        if ID not in label_json:
            return None, {STATUS: "unexpected error"}
        return DetectionLabel(self.token, self.endpoint, label_json), RESULT_OK

    def create_object(self, label_id, image_id, data, meta_data=None):
        """
        Create detection object on some image with some label and coordinates.
        :param label_id: id of detection label
        :param image_id: id of image
        :param data: [xmin, ymin, xmax, ymax] represent bounding box
        :param meta_data: json/dict of additional meta data
        :return: DetectionObject
        """
        label_json = self.post(
            OBJECT_ENDPOINT, data={DETECTION_LABEL: label_id, IMAGE: image_id, DATA: data, META_DATA: meta_data}
        )
        if ID not in label_json:
            return None, {STATUS: "unexpected error"}
        return DetectionObject(self.token, self.endpoint, label_json), RESULT_OK

    def upload_images(self, records):
        """
        Upload one or more files and add objects associated with them.
        :param records: list of dictionaries with objects and one of '_base64', '_file', '_url'
                        specify noresize: True to save image without (default False)
                        [
                            {
                                '_file': '__FILE_PATH__',
                                'objects': [
                                    'detection_label': '__UUID__'
                                    'data': [__xmin__, __ymin__, __xmax__, __ymax__],
                                    'metadata': {__metadata__}
                                ],
                                'noresize': False,
                             },
                             ...
                        ]
        :return: image, status
        """
        images = []
        worst_status = RESULT_OK
        for record in records:
            files, data = None, None
            noresize = NORESIZE in record and record[NORESIZE]
            noresize_on_server = noresize or self.max_image_size > 1024
            metadata = record[META_DATA] if META_DATA in record and record[META_DATA] else {}
            test_image = record[TEST_IMAGE] if TEST_IMAGE in record else False

            data = self._create_image_data(record, noresize, noresize_on_server, test_image, metadata)

            image_json = self.post(IMAGE_ENDPOINT, files=files, data=data)

            if image_json is None:
                worst_status = {STATUS: "image not uploaded " + str(record)}
                continue
            elif ID not in image_json:
                worst_status = {STATUS: "image not uploaded " + str(record)}
                continue

            image = Image(self.token, self.endpoint, image_json)

            if OBJECTS in record:
                for object in record[OBJECTS]:
                    self.create_object(object[DETECTION_LABEL], image.id, object[DATA], object[META_DATA])

            images.append(image)
        return images, worst_status

    def add_label_to_object(self, object_id, label_id, value=None):
        """
        Add recognition label to the object.
        :param label_id: id (uuid) of label
        :return: result
        """
        data = {LABEL_ID: label_id} if value is None else {LABEL_ID: label_id, "value": value}
        return self.post(OBJECT_ENDPOINT + object_id + "/add-label/", data=data)

    def construct_data(self, records=[], task_id=None, version=None, keep_prob=None, store_images=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")

        if task_id is None:
            raise Exception("Please specify task")

        data = {RECORDS: self.preprocess_records(records), TASK_ID: task_id, VERSION: version}

        if store_images:
            data[STORE_IMAGES] = True

        if keep_prob:
            data["keep_prob"] = keep_prob

        return data

    def detect_on_task(self, records=[], task_id=None, version=None, keep_prob=None, store_images=None):
        """
        Takes the images and calls the ximilar client for detection these images on the task.

        Usage:
            client = DetectionClient('__YOUR_API_TOKEN__')
            result = client.detect_on_task({'_url':'__SOME_IMG_URL__'}, task_id="__UUID__")

        :param records: array of json/dicts [{'_url':'url-path'}, {'_file': ''}, {'_base64': 'base64encodeimg'}]
        :param task_id: id of task
        :param version: optional(integer of specific version), default None/production_version
        :return: json response
        """
        # version is default set to None, so ximilar will determine which one to take
        data = self.construct_data(
            records=records, task_id=task_id, version=version, keep_prob=keep_prob, store_images=store_images
        )
        result = self.post(self.PREDICT_ENDPOINT, data=data)

        self.check_json_status(result)
        return result


class DetectionTask(DetectionClient):
    def __init__(self, token, endpoint, task_json, max_image_size):
        super().__init__(token, endpoint, max_image_size=max_image_size, resource_name=None)

        self.id = task_json[ID]
        self.name = task_json[NAME]
        self.description = task_json[DESCRIPTION] if DESCRIPTION in task_json else ""
        self.workspace = task_json[WORKSPACE] if WORKSPACE in task_json else DEFAULT_WORKSPACE

    def train(self):
        """
        Create new training/model and add it to the queue.
        :return: None
        """
        return self.post(TASK_ENDPOINT + self.id + "/train/")

    def remove(self):
        """
        Removes Detection Task.
        """
        self.remove_task(self.id)

    def add_label(self, label_id):
        """
        Add detection label to this task.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id + "/add-label/", data={LABEL_ID: label_id})

    def detach_label(self, label_id):
        """
        Remove/Detach detection label from the task.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id + "/remove-label/", data={LABEL_ID: label_id})

    def get_labels(self):
        """
        Get labels of this task.
        :return: list of Labels
        """
        if LABELS in self.cache:
            return self.cache[LABELS], RESULT_OK
        else:
            labels, result = self.get_all_labels(suffix="?task=" + self.id)

            if result[STATUS] == STATUS_OK:
                self.cache[LABELS] = labels

            return self.cache[LABELS], result

    def add_negative_image(self, image_id):
        """
        Add negative image to the detection task
        :param image_id: negative image ID to link with detection task
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id + "/add-image/", data={IMAGE_ID: image_id})

    def detect(self, records, version=None):
        """
        Takes the images and calls the ximilar client for detecting these images on the task.

        Usage:
            client = DetectionClient('__YOUR_API_TOKEN__')
            task = client.get_task('__TASK_ID__')
            result = task.detect({'_url':'__URL__'})

        :param records: array of json/dicts [{'_url':'url-path'}, {'_file': ''}, {'_base64': 'base64encodeimg'}]
        :param version: optional(integer of specific version), default None/production_version
        :return: json response
        """
        records = self.preprocess_records(records)

        # version is default set to None, so ximilar will determine which one to take
        data = {RECORDS: records, TASK_ID: self.id, VERSION: version}
        return self.post(DETECT_ENDPOINT, data=data)

    def to_json(self):
        labels, status = self.get_labels()
        return {TASK_ID: self.id, NAME: self.name, LABELS: [label.id for label in labels]}


class DetectionLabel(DetectionClient):
    """
    DetectionLabel entity from /detection/v2/label endpoint.
    DetectionLabel is connected to DetectionTasks and can also have Recognition Tasks.
    """

    def __init__(self, token, endpoint, label_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = label_json[ID]
        self.name = label_json[NAME]
        self.description = label_json[DESCRIPTION] if DESCRIPTION in label_json else ""
        self.workspace = label_json[WORKSPACE] if WORKSPACE in label_json else DEFAULT_WORKSPACE
        self.recognition_tasks = label_json[RECOGNITION_TASKS] if RECOGNITION_TASKS in label_json else None
        self.color = label_json[COLOR]
        self.output_name = (
            label_json[OUTPUT_NAME] if OUTPUT_NAME in label_json and label_json[OUTPUT_NAME] else self.name
        )

    def __str__(self):
        return self.name

    def remove(self):
        """
        Removes detection label.
        """
        return self.remove_label(self.id)

    def get_images(self):
        images, status = self.get_all_paginated_items(IMAGE_ENDPOINT + "?detection_labels=" + self.id)
        if not images and status[STATUS] == STATUS_ERROR:
            return None, status
        return [Image(self.token, self.endpoint, image) for image in images], RESULT_OK

    def get_training_images(self, page_url=None, verification=None, test=False):
        """
        Get paginated result of images for specific label.

        :param page_url: optional, select the specific page of images, default first page
        :return: (list of images, next_page)
        """
        if page_url is None:
            page_url = IMAGE_ENDPOINT + "?detection_labels=" + self.id

        if test:
            page_url += "&test=true"

        return super().get_training_images(page_url=page_url, verification=verification)

    def add_recognition_task(self, task_id):
        """
        Add recognition task to this label.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(LABEL_ENDPOINT + self.id + "/add-task/", data={TASK_ID: task_id})

    def detach_recognition_task(self, task_id):
        """
        Remove/Detach recognition task from label.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(LABEL_ENDPOINT + self.id + "/remove-task/", data={TASK_ID: task_id})

    def to_json(self):
        return {LABEL_ID: self.id, NAME: self.name, OUTPUT_NAME: self.output_name, COLOR: self.color}


class DetectionObject(DetectionClient):
    """
    Object/Bounding Box entity from /detection/v2/object endpoint.
    Every Object is located on some image and represents some detection label with coordinates (data).
    Coordinates are [xmin, ymin, xmax, ymax].
    Every object can also contain recognition labels.
    """

    def __init__(self, token, endpoint, object_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = object_json[ID]
        self.image = object_json[IMAGE]
        self.detection_label = object_json[DETECTION_LABEL]
        self.data = object_json[DATA]
        self.recognition_labels = object_json[RECOGNITION_LABELS] if RECOGNITION_LABELS in object_json else None

        # if the meta data was downloaded from the server, this field is never None
        self.meta_data = (
            None if META_DATA not in object_json else (object_json[META_DATA] if object_json[META_DATA] else {})
        )

        self._image_data = None  # The full image not actual object

    def remove(self):
        """
        Removes detection object.
        """
        return self.remove_object(self.id)

    def change_label(self, detection_label, data=None):
        bbox = data if data is not None else self.data
        self.data = bbox

        return self.put(
            OBJECT_ENDPOINT + self.id, data={"detection_label": detection_label, "image": self.image, "data": bbox}
        )

    def update_label(self, label_id, value=None):
        """
        Update value of the label on object.
        """
        data = {LABEL_ID: label_id, "value": value}
        return self.post(OBJECT_ENDPOINT + self.id + "/update-label/", data=data)

    def add_recognition_label(self, label_id, value=None):
        """
        Add recognition label to the object.
        :param label_id: id (uuid) of label
        :return: result
        """
        data = {LABEL_ID: label_id} if value is None else {LABEL_ID: label_id, "value": value}
        return self.post(OBJECT_ENDPOINT + self.id + "/add-label/", data=data)

    def detach_recognition_label(self, label_id):
        """
        Detach recognition label from the object.
        :param label_id: id (uuid) of label
        :return: result
        """
        return self.post(OBJECT_ENDPOINT + self.id + "/remove-label/", data={LABEL_ID: label_id})

    def _ensure_meta_data(self):
        """
        If the meta_data were not downloaded yet, do so
        """
        if not self.meta_data or not self.recognition_labels:
            data = self.get(OBJECT_ENDPOINT + self.id)
            self.meta_data = data[META_DATA]
            self.recognition_labels = data[RECOGNITION_LABELS]
        if not self.meta_data:
            self.meta_data = {}

    def get_meta_data(self):
        """
        Return the image meta data (dictionary) or empty dictionary
        :return: None is never returned
        """
        self._ensure_meta_data()
        return self.meta_data

    def add_meta_data(self, meta_data):
        """
        Add some meta data to image (extends already present meta data).
        """
        self._ensure_meta_data()
        if meta_data is None or not isinstance(meta_data, dict):
            raise Exception("Please specify dictionary of meta_data as param!")

        new_data = dict(list(self.meta_data.items()) + list(meta_data.items()))
        result = self.put(
            OBJECT_ENDPOINT + self.id,
            data={DATA: self.data, DETECTION_LABEL: self.detection_label[ID], META_DATA: new_data},
        )
        self.meta_data = result[META_DATA]
        return True

    def clear_meta_data(self):
        """
        Clear all meta data of image.
        """
        result = self.put(
            OBJECT_ENDPOINT + self.id, data={DATA: self.data, DETECTION_LABEL: self.detection_label[ID], META_DATA: {}}
        )
        self.meta_data = result[META_DATA]
        return True

    def to_json(self):
        self._ensure_meta_data()
        return {
            IMAGE: self.image,
            ID: self.id,
            DETECTION_LABEL: self.detection_label,
            DATA: self.data,
            LABELS: [
                {"id": label["id"], "name": label["name"], "value": label.get("value", None)}
                for label in self.recognition_labels
            ],
            META_DATA: self.meta_data,
        }

    def to_record(self):
        """Returns the object as a record."""
        return {
            ID: self.id,
            "bound_box": self.data,
            "name": self.detection_label["name"],
        }

    def extract_object_image(self):
        import cv2
        import numpy

        xmin, ymin, xmax, ymax = self.data
        response = urlopen(self.image[IMG_PATH])
        arr = numpy.asarray(bytearray(response.read()), dtype="uint8")
        img = cv2.imdecode(arr, cv2.IMREAD_UNCHANGED)
        return img[ymin:ymax, xmin:xmax]

    def get_bbox(self):
        return self.data

    def __str__(self):
        if isinstance(self.detection_label, dict):
            return self.id + " " + self.detection_label["name"]
        return self.id + " " + self.detection_label



from ximilar.client import RestClient
from ximilar.client.constants import (
    COLOR_SIMILARITY,
    RECORDS,
    ENDPOINT,
    FIELDS_TO_RETURN,
    _ID,
    FILTER,
    COUNT,
    K_COUNT,
    QUERY_RECORD,
    COLLECTION_ID,
    PHOTO_SIMILARITY,
    PRODUCT_SIMILARITY,
    FASHION_SIMILARITY,
    FASHION_SIMPLE,
    HOMEDECOR_SIMILARITY,
    CUSTOM_SIMILARITY,
    IMAGE_MATCHING,
    ANSWER_RECORDS,
    NEXT,
)

SIMILARITY_PHOTOS_DEPRECATED = "similarity/photos/v2/"
SIMILARITY_PHOTOS = "photo/search/v2/"
SIMILARITY_PRODUCTS = "similarity/products/v2/"
SIMILARITY_FASHION = "similarity/fashion/v2/"
SIMILARITY_FASHION_SIMPLE = "similarity/fashion-simple/v2/"
SIMILARITY_CUSTOM = "similarity/custom/v2/"
SIMILARITY_COLORS = "similarity/colors/v2/"
SIMILARITY_HOMEDECOR = "similarity/homedecor/v2/"
IMAGE_MATCHING_API = "image_matching/v2/"

SEARCH_OBJ_ENDPOINT = "search_by_object"
SEARCH = "search"
DETECT = "detect"
INSERT = "insert"
DELETE = "delete"
GET = "get"
UPDATE = "update"
RANDOM = "random"
KNN_VISUAL = "visualKNN"
KNN_VISUAL_TAGS = "visualTagsKNN"
PING = "ping"
RANK_RECORDS = "visualRankRecords"


class SimilarityPhotosClient(RestClient):
    def __init__(
        self, token, collection_id=None, endpoint=ENDPOINT + SIMILARITY_PHOTOS, resource_name=PHOTO_SIMILARITY
    ):
        super().__init__(token=token, endpoint=endpoint, max_image_size=512, resource_name=resource_name)
        self.headers[COLLECTION_ID] = collection_id
        self.PREDICT_ENDPOINT = KNN_VISUAL

    def construct_data(
        self, query_record=None, filter=None, k=5, fields_to_return=[_ID], custom_flow=None, records=[], **kwargs
    ):
        if query_record is None and len(records) == 0:
            raise Exception("Please specify record when using search method.")

        data = None
        if len(records):
            data = {RECORDS: self.preprocess_records(records), K_COUNT: k, FIELDS_TO_RETURN: fields_to_return}
        else:
            data = {
                QUERY_RECORD: self.preprocess_records([query_record])[0],
                K_COUNT: k,
                FIELDS_TO_RETURN: fields_to_return,
            }

        if filter:
            data[FILTER] = filter

        if kwargs:
            data.update(kwargs)

        data = self.fill_data(data, custom_flow)
        return data

    def fill_data(self, data, custom_flow):
        if custom_flow is not None:
            data["custom_flow"] = custom_flow

        return data

    def allRecords(self, size=1000, page=1, fields_to_return=[_ID], filterf={}):
        if size > 0:
            result = self.post("allRecords?size=%s&page=%s" % (size, page), data={FIELDS_TO_RETURN: fields_to_return})
        else:
            result = self.post("allRecords", data={FIELDS_TO_RETURN: fields_to_return, FILTER: filterf})
        return result

    def get_all_ids(self, fields=[], filterf={}):
        """
        Returns an array with records with "_id" fields.
        :return: {"answer_records": [an array with records with "_id" fields], "answer_count": "# of records"}
        """
        return self.allRecords(size=0, fields_to_return=[_ID] + fields, filterf=filterf)

    def all_records_iter(self, fields_to_return=[_ID], batch_size=1000, return_batch=False):
        """
        Get iterator over all records in the collection; the requests are done using batches & pagination.
        :param batch_size: optional, integer saying the batch size to retrieve the records with
        :param fields_to_return fields to be returned for each record in the collection ("_id" is returned always)
        :param return_batch: optional, boolean saying whether to return the batch or iterate over the records
        :return: iterator over records with the fields specified
        """
        page_counter = 1
        result = self.allRecords(batch_size, page_counter, fields_to_return)
        while ANSWER_RECORDS in result and len(result[ANSWER_RECORDS]) > 0:
            if return_batch:
                yield result[ANSWER_RECORDS]
            else:
                for rec in result[ANSWER_RECORDS]:
                    yield rec
            if NEXT not in result:
                break
            page_counter += 1
            result = self.allRecords(batch_size, page_counter, fields_to_return)

    def search(self, query_record, filter=None, k=5, fields_to_return=[_ID]):
        """
        Calls visual knn
        :param query_record: dictionary with field '_id' (from your collection) or '_url' or "_base64' data
        :param k: how many similar items to return
        :param fields_to_return: fields to return in every record
        :param filter: how to filter picked items (mongodb syntax)
        :return: json response
        """
        data = self.construct_data(query_record, filter=filter, k=k, fields_to_return=fields_to_return)
        return self.post(self.PREDICT_ENDPOINT, data=data)

    def search_and_rank(self, query_record, records, fields_to_return=[_ID]):
        """
        Ranks the records agains query
        :param query_record: dictionary with field '_id' (from your collection) or '_url' or "_base64' data
        :param k: how many similar items to return
        :param fields_to_return: fields to return in every record
        :param filter: how to filter picked items (mongodb syntax)
        :return: json response
        """
        data = self.construct_data(query_record, fields_to_return=fields_to_return)
        data[RECORDS] = self.preprocess_records(records)
        del data[K_COUNT]
        return self.post(RANK_RECORDS, data=data)

    def random(self, filter=None, count=10, fields_to_return=[_ID]):
        """
        Call random endpoint which returns random(X=count) records from your collection.
        :param filter: how to filter picked items (mongodb syntax)
        :param count: number of items to return
        :param fields_to_return: fields to return in every record
        :return: json response
        """
        data = {COUNT: count, FIELDS_TO_RETURN: fields_to_return}
        if filter:
            data[FILTER] = filter

        return self.post(RANDOM, data=data)

    def update(self, records, fields_to_return=["*"]):
        """
        Update records with meta-information, this will not update the descriptor(please do delete and insert instead).
        :param records: list of dictionaries with field '_id', '_url' or '_base64' with your meta-info
        :param fields_to_return: fields to return in every record
        :return: json response
        """
        data = {RECORDS: records, FIELDS_TO_RETURN: fields_to_return}
        return self.post(UPDATE, data=data)

    def remove(self, records):
        """
        Delete records from your collection.
        :param records: list of dictionaries with '_id' field
        :return: json response
        """
        data = {RECORDS: records}
        return self.post(DELETE, data=data)

    def insert(self, records, request_fields={}):
        """
        Insert records into collection with all meta information.
        :param records: dictionary with your "_id" and with one of "_url", "_file" or "_base64" to extract descriptor.
        :return: json response
        """
        records = self.preprocess_records(records)
        data = {RECORDS: records}
        if request_fields:
            data.update(request_fields)
        return self.post(INSERT, data=data)

    def get_records(self, records, fields_to_return=[_ID]):
        """
        Get the records from your collection.
        :param records: list of dictionaries with "_id" (identification of your records)
        :param fields_to_return: fields to return in every record
        :return: json response
        """
        data = {RECORDS: records, FIELDS_TO_RETURN: fields_to_return}
        return self.post(GET, data=data)

    def ping(self):
        return self.post(PING)


class SimilarityProductsClient(SimilarityPhotosClient):
    def __init__(
        self, token, collection_id=None, endpoint=ENDPOINT + SIMILARITY_PRODUCTS, resource_name=PRODUCT_SIMILARITY
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)


class SimilarityCustomClient(SimilarityPhotosClient):
    def __init__(
        self, token, collection_id=None, endpoint=ENDPOINT + SIMILARITY_CUSTOM, resource_name=CUSTOM_SIMILARITY
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)


class SimilarityColorsClient(SimilarityPhotosClient):
    def __init__(
        self, token, collection_id=None, endpoint=ENDPOINT + SIMILARITY_COLORS, resource_name=COLOR_SIMILARITY
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)


class ImageMatchingSearchClient(SimilarityPhotosClient):
    def __init__(self, token, collection_id=None, endpoint=ENDPOINT + IMAGE_MATCHING_API, resource_name=IMAGE_MATCHING):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)


class SimilarityHomeDecorClient(SimilarityPhotosClient):
    def __init__(
        self, token, collection_id=None, endpoint=ENDPOINT + SIMILARITY_HOMEDECOR, resource_name=HOMEDECOR_SIMILARITY
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)


class SimilarityFashionClient(SimilarityPhotosClient):
    def __init__(
        self,
        token,
        collection_id=None,
        endpoint=ENDPOINT + SIMILARITY_FASHION,
        resource_name=FASHION_SIMILARITY,
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = KNN_VISUAL_TAGS

    def insert(self, records, custom_flow=None, skip_empty=False):
        """
        Insert records into collection with all meta information.
        :param custom_flow: string ID of the flow that should be called during the insert (in extractor)
        :param records: dictionary with your "_id" and with one of "_url", "_file" or "_base64" to extract descriptor.
        :param skip_empty: boolean, if True, empty records will be skipped (default is False)
        :return: json response
        """
        records = self.preprocess_records(records)
        data = {RECORDS: self.fill_data(records, custom_flow), "skip_empty": skip_empty}
        return self.post(INSERT, data=data)

    def descriptor(self, records, custom_flow=None, **kwargs):
        data = self.construct_data(records=records, custom_flow=custom_flow, **kwargs)
        return self.post("descriptor", data=data)


class SimilarityFashionSimpleClient(SimilarityFashionClient):
    def __init__(
        self,
        token,
        collection_id=None,
        endpoint=ENDPOINT + SIMILARITY_FASHION_SIMPLE,
        resource_name=FASHION_SIMPLE,
    ):
        super().__init__(token=token, collection_id=collection_id, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = KNN_VISUAL



import os
import hashlib

from ximilar.client.constants import *
from ximilar.client import RecognitionClient


PRODUCT_ENDPOINT = "product/v2/product/"
IMAGE_ENDPOINT = "product/v2/image/"
COLLECTION_ENDPOINT = "product/v2/collection/"


class ProductClient(RecognitionClient):
    """
    Ximilar API Client for Products management.
    """

    def __init__(
        self, token, endpoint=ENDPOINT, workspace=DEFAULT_WORKSPACE, max_image_size=0, resource_name=PRODUCTS_MANAGEMENT
    ):
        super().__init__(
            token=token,
            endpoint=endpoint,
            workspace=workspace,
            max_image_size=max_image_size,
            resource_name=resource_name,
        )

    def create_product(self, name, customer_product_id, product_collection, meta_data=None):
        data = {NAME: name, "customer_product_id": customer_product_id, "product_collection": product_collection}

        if meta_data:
            data[META_DATA] = meta_data

        product_json = self.post(PRODUCT_ENDPOINT, data=data, params=self.add_workspace(None, None))
        if ID not in product_json:
            return None, {STATUS: "unexpected error"}
        return Product(self.token, self.endpoint, self.workspace, product_json), RESULT_OK

    def create_image(self, record, name, product, meta_data={}, md5s=[]):
        assert product

        record = self._create_image_data(record, True, True, False, meta_data)

        record[NAME] = name
        record["product"] = product

        if "base64" in record and len(md5s):
            md5h = hashlib.md5(record["base64"].encode("utf-8")).hexdigest()
            if md5h in md5s:
                return None, {STATUS: "skipped"}

        image_json = self.post(IMAGE_ENDPOINT, data=record)
        if ID not in image_json:
            return None, {STATUS: "unexpected error"}
        return ProductImage(self.token, self.endpoint, self.workspace, image_json), RESULT_OK

    def get_product(self, product_id):
        product_json = self.get(PRODUCT_ENDPOINT + product_id)
        if ID not in product_json:
            return None, {STATUS: "Product with this id not found!"}
        return Product(self.token, self.endpoint, self.workspace, product_json), RESULT_OK

    def get_image(self, image_id):
        image_json = self.get(IMAGE_ENDPOINT + image_id)
        if ID not in image_json:
            return None, {STATUS: "Image with this id not found!"}
        return ProductImage(self.token, self.endpoint, self.workspace, image_json), RESULT_OK

    def remove_product(self, product_id):
        return self.delete(PRODUCT_ENDPOINT + product_id)

    def remove_image(self, image_id):
        return self.delete(IMAGE_ENDPOINT + image_id)

    def get_images_url(self, page_url, product=None, name=None):
        url = (
            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
            if page_url
            else IMAGE_ENDPOINT + "?page=1"
        )
        if product is not None:
            url = url + "&product=" + str(product)
        if name is not None:
            url = url + "&name=" + str(name)
        return url

    def get_all_products(
        self, page_url=None, product_collection=None, name=None, customer_product_id=None, collection_type=None
    ):
        url = self.get_products_url(page_url, product_collection, name, customer_product_id, collection_type)

        products, status = self.get_all_paginated_items(url)
        if not products and status[STATUS] == STATUS_ERROR:
            return None, status
        return [Product(self.token, self.endpoint, self.workspace, p_json) for p_json in products], RESULT_OK

    def get_all_images(self, product=None):
        url = IMAGE_ENDPOINT
        if product is not None:
            url = url + "?product=" + str(product)

        images, status = self.get_all_paginated_items(url)
        if not images and status[STATUS] == STATUS_ERROR:
            return None, status

        return [ProductImage(self.token, self.endpoint, self.workspace, i_json) for i_json in images], RESULT_OK

    def get_images_by_name(self, page_url, name=None):
        url = self.get_images_url(page_url, name=name)
        result = self.get(url)
        return (
            [ProductImage(self.token, self.endpoint, self.workspace, i_json) for i_json in result[RESULTS]],
            result[NEXT],
            {"count": result["count"], STATUS: "ok"},
        )

    def get_products_url(
        self,
        page_url=None,
        product_collection=None,
        name=None,
        customer_product_id=None,
        collection_type=None,
        ordering=None,
        empty=None,
    ):
        url = (
            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
            if page_url
            else PRODUCT_ENDPOINT + "?page=1"
        )

        if page_url is not None:
            return url
        if name is not None:
            url += "&search=" + str(name)
        if product_collection is not None:
            url += "&product_collection=" + str(product_collection)
        if customer_product_id is not None:
            url += "&customer_product_id=" + str(customer_product_id)
        if collection_type is not None:
            url += "&collection_type=" + str(collection_type)
        if ordering is not None:
            url += "&ordering=" + str(ordering)
        if empty is not None:
            url += "&empty=true"
        return url

    def get_products(
        self,
        page_url=None,
        product_collection=None,
        name=None,
        customer_product_id=None,
        collection_type=None,
        ordering=None,
        empty=None,
    ):
        url = self.get_products_url(
            page_url, product_collection, name, customer_product_id, collection_type, ordering, empty
        )
        result = self.get(url)
        return (
            [Product(self.token, self.endpoint, self.workspace, product_json) for product_json in result[RESULTS]],
            result[NEXT],
            {"count": result["count"], STATUS: "ok"},
        )

    def get_product_by_customer_id(self, customer_product_id):
        products, next, status = self.get_products(customer_product_id=customer_product_id)
        if not products or len(products) == 0 or status[STATUS] == STATUS_ERROR:
            return None, status
        return products[0], RESULT_OK


class Product(ProductClient):
    def __init__(self, token, endpoint, workspace, product_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace, resource_name=None)

        self.id = product_json["id"]
        self.name = product_json["name"]
        self.product_collection = product_json["product_collection"]
        self.customer_product_id = product_json["customer_product_id"]
        self.meta_data = product_json.get("meta_data", None)
        self.thumb = product_json.get("thumb", None)

    def __str__(self):
        return self.name + ":" + self.id + ":" + str(self.product_collection)

    def remove(self):
        self.remove_product(self.id)

    def update_product(self, **kwargs):
        """Call this as following:

        product.update_product(
            **{
                "name": card.get("name", "") + " " + card.get("year", "") + " #" + card.get("card_number", ""),
                "customer_product_id": card["product"],
                "product_collection": exports[card["Subcategory"]]["id"],
                "meta_data": card
            }
        )
        """
        result = self.patch(PRODUCT_ENDPOINT + self.id, data=kwargs)

    def update_meta(self, new_data):
        result = self.patch(PRODUCT_ENDPOINT + self.id, data={META_DATA: new_data})

    def get_images(self):
        return self.get_all_images(str(self.id))


class ProductImage(ProductClient):
    def __init__(self, token, endpoint, workspace, image_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace, resource_name=None)

        self.id = image_json["id"]
        self.name = image_json["name"]
        self.product = image_json["product"]
        self.meta_data = image_json.get("meta_data", None)
        self.objects = image_json["_objects"]

        self.img = image_json["file"]
        self.img_thumb = image_json["thumb"]

    def __str__(self):
        product = self.product if self.product and isinstance(self.product, str) else ""
        return self.name + ":" + self.id + ":" + product

    def remove(self):
        self.remove_image(self.id)

    def refresh(self):
        newpi = self.get_image(self.id)
        self.name = newpi.name
        self.product = newpi.product
        self.meta_data = newpi.meta_data if newpi.meta_data is not None else {}
        self.objects = newpi.objects

    def update_meta(self, new_data, refresh=False):
        if refresh:
            self.refresh()

        if self.meta_data is None:
            self.meta_data = {}

        self.meta_data.update(new_data)
        self.patch(IMAGE_ENDPOINT + self.id, data={META_DATA: new_data})



from ximilar.client import RestClient
from ximilar.client.constants import *
from ximilar.client.constants import ENDPOINT

LLM_ENDPOINT = "llm/v2/llm"


class LLMClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, max_image_size=600, resource_name="", request_timeout=90):
        super().__init__(token, endpoint, max_image_size, resource_name, request_timeout)
        self.PREDICT_ENDPOINT = LLM_ENDPOINT

    def construct_data(self, records, **kwargs):
        if len(records) == 0:
            raise Exception("Please specify at least on record when using ocr endpoint.")

        data = {RECORDS: records}

        if kwargs:
            data.update(kwargs)

        return data

    def llm_response(self, records, lang="en", headers=None):
        """
        Call the language model endpoint.
        :param records: list of json records
        :param lang: specify the language of the text that is probably on image, defaults to "en" (currently only english is supported)
        :return: json result data from the API
        """
        data = self.construct_data(records, lang=lang)
        result = self.post(self.PREDICT_ENDPOINT, data=data, headers=headers)
        return result



import os

from ximilar.client.constants import *
from ximilar.client import RestClient
from ximilar.client.recognition import Image, RecognitionClient
from ximilar.client.detection import DetectionObject


TYPE_ENDPOINT = "similarity/training/v2/type/"
TASK_ENDPOINT = "similarity/training/v2/task/"
GROUP_ENDPOINT = "similarity/training/v2/group/"
DESCRIPTOR_ENDPOINT = "similarity/training/v2/descriptor"


class CustomSimilarityClient(RecognitionClient):
    """
    Ximilar API Client for Custom Similarity.
    """

    def __init__(
        self, token, endpoint=ENDPOINT, workspace=DEFAULT_WORKSPACE, max_image_size=512, resource_name=CUSTOM_SIMILARITY
    ):
        super().__init__(
            token=token,
            endpoint=endpoint,
            workspace=workspace,
            max_image_size=max_image_size,
            resource_name=resource_name,
        )
        self.PREDICT_ENDPOINT = DESCRIPTOR_ENDPOINT

    def create_group(self, name, description, sim_type, meta_data=None):
        data = {NAME: name, DESCRIPTION: description, "type": sim_type}

        if meta_data:
            data[META_DATA] = meta_data

        group_json = self.post(GROUP_ENDPOINT, data=data)
        if ID not in group_json:
            return None, {STATUS: "unexpected error"}
        return SimilarityGroup(self.token, self.endpoint, self.workspace, group_json), RESULT_OK

    def create_task(self, name, description=""):
        data = {NAME: name, DESCRIPTION: description}
        task_json = self.post(TASK_ENDPOINT, data=data)
        if ID not in task_json:
            return None, {STATUS: "unexpected error"}
        return SimilarityTask(self.token, self.endpoint, self.workspace, task_json), RESULT_OK

    def create_type(self, name, description=""):
        data = {NAME: name, DESCRIPTION: description}
        type_json = self.post(TYPE_ENDPOINT, data=data)
        if ID not in type_json:
            return None, {STATUS: "unexpected error"}
        return SimilarityType(self.token, self.endpoint, self.workspace, type_json), RESULT_OK

    def get_task(self, task_id):
        task_json = self.get(TASK_ENDPOINT + task_id)
        if ID not in task_json:
            return None, {STATUS: "SimilarityTask with this id not found!"}
        return SimilarityTask(self.token, self.endpoint, self.workspace, task_json), RESULT_OK

    def get_type(self, type_id):
        type_json = self.get(TYPE_ENDPOINT + type_id)
        if ID not in type_json:
            return None, {STATUS: "SimilarityType with this id not found!"}
        return SimilarityType(self.token, self.endpoint, self.workspace, type_json), RESULT_OK

    def get_group(self, group_id):
        group_json = self.get(GROUP_ENDPOINT + group_id)
        if ID not in group_json:
            return None, {STATUS: "SimilarityGroup with this id not found!"}
        return SimilarityGroup(self.token, self.endpoint, group_json["type"]["workspace"], group_json), RESULT_OK

    def remove_group(self, group_id):
        return self.delete(GROUP_ENDPOINT + group_id)

    def remove_type(self, type_id):
        return self.delete(TYPE_ENDPOINT + type_id)

    def remove_task(self, task_id):
        return self.delete(TASK_ENDPOINT + task_id)

    def get_all_tasks(self):
        tasks, status = self.get_all_paginated_items(TASK_ENDPOINT)
        if not tasks and status[STATUS] == STATUS_ERROR:
            return None, status
        return [SimilarityTask(self.token, self.endpoint, self.workspace, t_json) for t_json in tasks], RESULT_OK

    def get_all_types(self):
        types, status = self.get_all_paginated_items(TYPE_ENDPOINT)
        if not types and status[STATUS] == STATUS_ERROR:
            return None, status
        return [SimilarityType(self.token, self.endpoint, self.workspace, t_json) for t_json in types], RESULT_OK

    def get_groups_url(
        self,
        page_url=None,
        search=None,
        test=None,
        image=None,
        group=None,
        empty=None,
        type=None,
        object=None,
        count=None,
    ):
        url = (
            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
            if page_url
            else GROUP_ENDPOINT + "?page=1"
        )
        if page_url is None:
            url += "&" + search if search is not None else ""
        if test is not None:
            url += f"&test={'True' if test else 'False'}"
        if image is not None:
            url += "&image=" + str(image)
        if object is not None:
            url += "&object=" + str(object)
        if group is not None:
            url += "&group=" + str(group)
        if empty is not None:
            url += "&empty=true"
        if type is not None:
            url += "&type=" + str(type)
        if count is not None:
            url += "&count=" + str(count)
        return url

    def get_groups(
        self,
        page_url=None,
        search=None,
        test=None,
        image=None,
        group=None,
        empty=None,
        type=None,
        object=None,
        count=None,
    ):
        url = self.get_groups_url(page_url, search, test, image, group, empty, type, object, count)

        result = self.get(url)
        return (
            [SimilarityGroup(self.token, self.endpoint, self.workspace, group_json) for group_json in result[RESULTS]],
            result[NEXT],
            {"count": result["count"], STATUS: "ok"},
        )

    def get_all_groups(self, page_url=None, search=None, test=None, empty=None, type=None, count=None):
        url = self.get_groups_url(page_url, search, test, empty=empty, type=type, count=count)

        groups, status = self.get_all_paginated_items(url)
        return (
            [SimilarityGroup(self.token, self.endpoint, self.workspace, group_json) for group_json in groups],
            status,
        )

    def get_all_groups_by_name(self, name, sim_type=None, page_url=None, test=None):
        return self.get_all_groups(page_url, "search=" + name, test, type=sim_type)

    def get_group_by_name(self, name, sim_type=None, page_url=None, test=None):
        groups, _ = self.get_all_groups_by_name(name, sim_type=sim_type, page_url=page_url, test=test)
        for group in groups:
            if group.name == name:
                return group, RESULT_OK
        return None, None

    def get_all_groups_by_type(self, sim_type, page_url=None, test=None):
        return self.get_all_groups(page_url, sim_type, test, type=sim_type)

    def get_groups_by_type(self, sim_type, page_url=None, test=None):
        return self.get_groups(page_url, test=test, type=sim_type)

    def get_groups_by_type_name(self, type_name, page_url=None, test=None):
        return self.get_groups(page_url, "type__name=" + type_name, test=test)

    def descriptor(self, records, task_id, version=None):
        """
        Takes the images and calls the ximilar client for extracting visual descriptors.
        """
        # version is default set to None, so ximilar will determine which one to take
        data = self.construct_data(records=records, task_id=task_id, version=version)
        result = self.post(self.PREDICT_ENDPOINT, data=data)

        self.check_json_status(result)
        return result


class SimilarityTask(CustomSimilarityClient):
    def __init__(self, token, endpoint, workspace, task_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace, resource_name=None)

        self.id = task_json["id"]
        self.name = task_json["name"]

    def __str__(self):
        return self.name + ":" + self.id

    def add_type(self, type_id):
        return self.post(TASK_ENDPOINT + self.id + "/add-type/", data={TYPE_ID: type_id})

    def remove_type(self, type_id):
        return self.post(TASK_ENDPOINT + self.id + "/remove-type/", data={TYPE_ID: type_id})

    def train(self):
        pass

    def descriptor(self, json_records):
        return self.descriptor(json_records, self.id)

    def remove(self):
        self.remove_task(self.id)


class SimilarityType(CustomSimilarityClient):
    def __init__(self, token, endpoint, workspace, type_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace, resource_name=None)

        self.id = type_json["id"]
        self.name = type_json["name"]

    def __str__(self):
        return self.name + ":" + self.id

    def remove(self):
        self.remove_type(self.id)


class SimilarityGroup(CustomSimilarityClient):
    def __init__(self, token, endpoint, workspace, group_json):
        super().__init__(token, endpoint=endpoint, workspace=workspace, resource_name=None)

        self.id = group_json["id"]
        self.name = group_json["name"] if "name" in group_json else None
        self.verifyCount = 0

        self.description = group_json["description"] if "description" in group_json else None

        self.groups = None
        if "groups" in group_json:
            self.groups = [SimilarityGroup(token, endpoint, workspace, group) for group in group_json["groups"]]

        self.images = None
        if "images" in group_json:
            self.images = [Image(token, endpoint, image) for image in group_json["images"]]

        self.object = None
        if "detection_objects" in group_json:
            self.objects = [DetectionObject(token, endpoint, object_s) for object_s in group_json["detection_objects"]]

        self.meta_data = (
            None if META_DATA not in group_json else (group_json[META_DATA] if group_json[META_DATA] else {})
        )

        self.test_group = group_json["test_group"] if "test_group" in group_json else None
        if isinstance(group_json["type"], str):
            self.type = group_json["type"]
        elif isinstance(group_json["type"], dict):
            self.type = SimilarityType(token, endpoint, workspace, group_json["type"])

        self.type = group_json["type"] if "type" in group_json else None

    def __str__(self):
        return self.name + ":" + self.id

    def refresh(self, force):
        if force or self.images is None or self.groups is None:
            group, _ = self.get_group(self.id)

            self.workspace = group.workspace
            self.meta_data = group.meta_data
            self.name = group.name
            self.groups = group.groups
            self.objects = group.objects
            self.images = group.images
            self.type = group.type
            self.test_group = group.test_group
            self.description = group.description

    def get_images(self):
        self.refresh(False)
        return self.images

    def get_groups(self):
        self.refresh(False)
        return self.groups

    def add_images(self, images, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/add-images/", data={"images": images})
        self.refresh(refresh)
        return result

    def add_objects(self, objects, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/add-objects/", data={"objects": objects})
        self.refresh(refresh)
        return result

    def remove_images(self, images, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/remove-images/", data={"images": images})
        self.refresh(refresh)
        return result

    def remove_objects(self, objects, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/remove-objects/", data={"objects": objects})
        self.refresh(refresh)
        return result

    def add_groups(self, groups, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/add-groups/", data={"groups": groups})
        self.refresh(refresh)
        return result

    def remove_groups(self, groups, refresh=False):
        result = self.post(GROUP_ENDPOINT + self.id + "/remove-groups/", data={"groups": groups})
        self.refresh(refresh)
        return result

    def update_stype(self, stype):
        result = self.put(GROUP_ENDPOINT + self.id, data={"type": stype})
        return result

    def update_name(self, name, stype):
        result = self.put(GROUP_ENDPOINT + self.id, data={"name": name, "type": stype})
        return result

    def set_test(self, test, refresh=False):
        field = "mark-test" if test else "unmark-test"
        result = self.post(GROUP_ENDPOINT + "update", data={"groups": [self.id], field: True})
        self.refresh(refresh)
        return result

    def remove(self):
        self.remove_group(self.id)

    def _ensure_meta_data(self):
        """
        If the meta_data were not downloaded yet, do so
        """
        if self.meta_data is None:
            self.refresh(True)

    def get_meta_data(self):
        """
        Return the group meta data (dictionary) or empty dictionary.
        :return: None is never returned
        """
        self._ensure_meta_data()
        return self.meta_data

    def add_meta_data(self, meta_data):
        """
        Add some meta data to group (extends already present meta data).
        """
        self._ensure_meta_data()
        if meta_data is None or not isinstance(meta_data, dict):
            raise Exception("Please specify dictionary of meta_data as param!")

        new_data = dict(list(self.meta_data.items()) + list(meta_data.items()))
        result = self.patch(GROUP_ENDPOINT + self.id, data={META_DATA: new_data})
        self.meta_data = result[META_DATA]
        return True

    def _ensure_verifications(self):
        json_results = self.get("annotate/v2/similarity-verification/?group=" + self.id)
        self.verifyCount = len(json_results["results"])
        return json_results["results"]

    def verify(self, user_id):
        """
        Verify this group by some user.
        """
        result = self.post("annotate/v2/similarity-verification/", data={USER: user_id, "group_id": self.id})
        self._ensure_verifications()
        return result

    def unverify(self):
        """
        Unverify all users from group.
        """
        results = self._ensure_verifications()
        for result in results:
            self.delete("annotate/v2/similarity-verification/" + str(result["id"]))
        self._ensure_verifications()
        return True

    def download(self, destination):
        os.makedirs(os.path.join(destination, self.name), exist_ok=True)

        for image in self.images:
            # print("image:,", image, type(image))
            # image = Image(self.token, self.endpoint, image)
            image.download(os.path.join(destination, self.name))

        for object_1 in self.objects:
            image = object_1.image
            image = Image(self.token, self.endpoint, image)
            image.download(os.path.join(destination, self.name))


class SimilarityRunLogClient(RestClient):
    BASE_ENDPOINT = "/similarity/v2/run-log/"

    def __init__(self, token, endpoint=ENDPOINT):
        self.workspace = None
        super().__init__(token=token, endpoint=endpoint, resource_name=None)

    def create_log(self, collection, data):
        return self.post(self.BASE_ENDPOINT, {"collection": collection, "data": data})

    def get_log(self, log_id):
        return self.get(self.BASE_ENDPOINT + log_id)

    def delete_log(self, log_id):
        return self.delete(self.BASE_ENDPOINT + log_id)

    def update_log(self, log_id, data):
        return self.put(self.BASE_ENDPOINT + log_id, {"data": data})



import gzip
import json
import os
import string
import sys
from os.path import splitext


def read_json_array(file_name, required_fields=None, skip=0):
    with open(file_name, "r", encoding="utf-8") as f:
        array = json.load(f)
        for record in array:
            if skip > 0:
                skip -= 1
                continue
            if not required_fields or all(field in record for field in required_fields):
                yield record


def read_json_file_iterator(file_name, required_fields=None, skip=0):
    """
    This iterator method reads a file with JSON formatted lines, checks that every line has given field(s) and
    returns an iterator over JSON objects constructed from these lines.
    :param file_name: file with JSON formatted lines
    :param required_fields: list of fields that should be in each JSON object on the lines; lines not matching are not
            returned
    :return: an iterator over JSON objects that have given required fields
    """
    if splitext(file_name)[1] == ".gz":
        f = gzip.open(file_name, "rt", encoding="utf-8")
    else:
        f = open(file_name, "r", encoding="utf-8")

    chars_to_strip = string.whitespace + ","
    for line in f:
        line = line.rstrip(chars_to_strip)
        if skip > 0:
            skip -= 1
            continue
        try:
            record = json.loads(line)
        except UnicodeDecodeError as e:
            print("wrong encoding: '" + line + "'")
            continue
        except ValueError as e:
            print("JSON error: '" + line + "'")
            print(e)
            continue

        if not required_fields or all(field in record for field in required_fields):
            yield record
    f.close()


def read_json_file_list(file_name, required_fields=None, skip=0, is_array=False):
    data = []
    iterator = (
        read_json_array(file_name, required_fields, skip)
        if is_array
        else read_json_file_iterator(file_name, required_fields, skip)
    )
    for record in iterator:
        data.append(record)
    return data


class JSONWriter(object):
    """
    This class encapsulates a writer of JSON records into a potentially GZIPed file.
    """

    def __init__(self, file_name):
        if not file_name:
            self.output = sys.stdout
        else:
            if splitext(file_name)[1] == ".gz":
                self.output = gzip.open(file_name, "w")
            else:
                self.output = open(file_name, "w", encoding="utf-8")

    def __enter__(self):  # this is necessary to be used in "with statement"
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            print(f"error processing JSON writer {str(exc_type)}, {str(exc_value)}, {str(traceback)}", file=sys.stderr)
        return self.close()

    def write(self, json_record, ensure_ascii=False, indent=None):
        json.dump(json_record, self.output, ensure_ascii=ensure_ascii, indent=indent)
        self.output.write("\n")
        self.output.flush()

    def close(self):
        if self.output is not sys.stdout:
            self.output.close()


def all_values(json_rec, field_name, subfield_name=None):
    """
    This is a helper function that, given a field name, returns an iterator over all values in the field regardless of
    if the field contains an array or a single value.
    :param json_rec: JSON record
    :param field_name: field to contain either array or a single value
    :return: iterator over the values
    """
    if field_name not in json_rec:
        return

    field = json_rec[field_name][subfield_name] if subfield_name else json_rec[field_name]

    if isinstance(field, list):
        for k in field:
            yield k
    else:
        yield field


def traverse_dir_to_json(output_file, directory):
    """
    Traverse a directory recursively and, for each file in the tree, add a single JSON line into an output JSON file:
    { "_file": "path/excluding/the/directory/file.ext" }
    :param output_file: JSON file to generate (it's overwritten if exists)
    :param directory: directory to traverse recursively
    :return: true, if the file was generated alright
    """
    output = JSONWriter(output_file)
    try:
        for subdir, dirs, files in os.walk(directory):
            for file_name in files:
                output.write({"_file": os.path.join(subdir, file_name)})
    finally:
        output.close()



import requests
import json
import base64
import os
import re
import concurrent.futures
import urllib.parse

from tqdm import tqdm

from ximilar.client.constants import *
from ximilar.client.exceptions import XimilarClientException
from ximilar.client.utils.decorators import retry_when

CONFIG_ENDPOINT = "account/v2/config/"
BASE64_HEADER_PATTERN = re.compile(r"^data:image/(\w+);base64,")


class RestClient(object):
    """
    Parent class that implements HTTP GET, POST, DELETE methods with requests lib and loading images to base64.

    All objects contains TOKEN and ENDPOINT information.
    """

    def __init__(self, token, endpoint=ENDPOINT, max_image_size=600, resource_name="", request_timeout=90):
        self.token = token
        self.cache = {}
        self.endpoint = endpoint
        self.max_image_size = max_image_size
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": self.get_token_header(self.token),
            "User-Agent": "Ximilar Client/Python: " + self.token[:5],
        }
        self.check_resource(resource_name)
        self.request_timeout = request_timeout

    def update_token(self, token):
        self.headers["Authorization"] = self.get_token_header(token)

    def add_workspace(self, data, url=None):
        """
        Add workspace uuid to the data.
        :param data: dictionary/json data which will be send to endpoint
        :return: modified json data with workspace
        """
        if self.workspace != DEFAULT_WORKSPACE:
            if data is None:
                data = {}

            # if workspace is already in url then do not create the param
            if url is not None and self.workspace not in url:
                data[WORKSPACE] = self.workspace
            elif url is None:
                data[WORKSPACE] = self.workspace
        return data

    def get_token_header(self, token):
        if len(token) < 70:
            return "Token " + token
        return "JWT " + token

    def invalidate(self):
        self.cache = {}

    def _type(self):
        """
        Returns the Class name of the object.
        For example RecogntionClient._type() => "RecognitionClient"
                    Image._type() => "Image"
        """
        return self.__class__.__name__

    @staticmethod
    def urljoin(*args):
        """
        Joins given arguments into an url. Trailing and leading slashes are
        stripped for each argument.
        """

        url = "/".join(map(lambda x: str(x).rstrip("/").lstrip("/"), args))
        return url

    @retry_when(ConnectionError)
    def get(self, api_endpoint, data=None, params=None):
        """
        Call the http GET request with data.
        :param api_endpoint: endpoint path
        :param data: optional data
        :param params: optional dictionary of URL params
        :return: json response
        """
        result = requests.get(
            self.urljoin(self.endpoint, api_endpoint),
            params=params,
            headers=self.headers,
            data=data,
            timeout=self.request_timeout,
        )
        return result.json()

    @retry_when(ConnectionError)
    def post(self, api_endpoint, data=None, files=None, params=None, method=requests.post, headers=None):
        """
        Call the http POST request with data.

        :param api_endpoint: endpoint path
        :param data: optional data
        :param files: optional files to upload
        :param params: optional dictionary of URL params
        :param method: POST or PUT method of request
        :return: json response
        """
        self.invalidate()
        if data is not None and files is not None:
            raise Exception("Unable to send data along with files with python requests library!")

        if data is not None:
            data = json.dumps(data)

        result = method(
            self.urljoin(self.endpoint, api_endpoint),
            params=params,
            headers=self.headers if headers is None else headers,
            data=data,
            files=files,
            timeout=self.request_timeout,
        )
        # todo: check JSON RESULT CODES -> raise XimilarClientException
        # todo: check HTTP STATUS CODES -> raise XimilarClientException
        try:
            json_result = result.json()
            return json_result
        except ValueError as e:
            return None

    @retry_when(ConnectionError)
    def put(self, api_endpoint, data=None, files=None, params=None):
        """
        Call the http PUT request with data
        """
        return self.post(api_endpoint, data=data, files=files, params=params, method=requests.put)

    @retry_when(ConnectionError)
    def patch(self, api_endpoint, data=None, files=None, params=None):
        """
        Call the http PATCH request with data
        """
        return self.post(api_endpoint, data=data, files=files, params=params, method=requests.patch)

    @retry_when(ConnectionError)
    def delete(self, api_endpoint, data=None, params=None):
        """
        Call the http DELETE request with data.
        :param api_endpoint: endpoint path
        :param data: optional data
        :param params: optional dictionary of URL params
        :return: response
        """
        self.invalidate()

        url = urllib.parse.urljoin(self.endpoint, api_endpoint)
        result = requests.delete(url, params=params, headers=self.headers, data=data, timeout=self.request_timeout)

        if result.status_code == HTTP_NO_CONTENT_204:
            return result

        return result.json()

    def check_resource(self, resource_name):
        """
        Checks if the user has access (resource) to the service.
        :param resource_name: name of the service
        :return: True if user has access otherwise False
        """
        # we don't want to check resource for entities like Image, Task, Object, DetectionLabel, ...
        if "localhost" in self.endpoint or "Client" not in self._type() or not self.token or not resource_name:
            return True

        # we need to authorize it with FIXED Endpoint https://api.ximilar.com/authorization/v2/authorize
        # as the self.endpoint can be different size
        result = requests.post(
            ENDPOINT + "authorization/v2/authorize",
            data=json.dumps({"service": resource_name, "call_type": "authenticate"}),
            headers=self.headers,
            timeout=10,
        )

        try:
            result = result.json()
        except:
            result = None

        if result and USER_ID in result:
            return True

        if "detail" in result:
            raise Exception(result["detail"] + " " + resource_name + ". Please contact tech@ximilar.com!")
        raise Exception("User has no access for service: " + resource_name + ". Please contact tech@ximilar.com!")

    def get_user_details(self):
        """
        Get Basic information of actual user.
        :return: json response if success
        """
        return self.get("account/v2/user/")

    def get_all_paginated_items(self, url):
        """
        Getting all paginated items from specific endpoint url
        :param url: url path which will be queried
        :return: items
        """
        count, items = 0, []

        while True:
            result = self.get(url)
            count = result.get("count", 0)
            if RESULTS in result:
                for item in result[RESULTS]:
                    items.append(item)
                if result[NEXT] is None:
                    break
            else:
                if DETAIL in result:
                    return None, {DETAIL: result[DETAIL], STATUS: STATUS_ERROR}
                else:
                    return None, RESULT_ERROR

            # we need to replace the full url just with the end, because endpoint
            # is automatically adde during the request
            url = (
                result[NEXT]
                .replace(self.endpoint, "")
                .replace(self.endpoint.replace("https", "http"), "")
                .replace("http://localhost/api/", "")
            )

        return items, {"status": "OK", "count": count}

    def add_header(self, header):
        """Add header to the every request"""
        self.headers.update(header)

    def check_json_status(self, result):
        """
        Check status code in json result, if not 200 then raises exception!
        :param result: json dictionary
        :raises XimilarClientException: returned status code from endpoint, text
        :raises XimilarClientException: 500, text
        """
        if result is not None and "status" in result:
            if "code" in result["status"] and result["status"]["code"] != 200:
                if result["status"]["text"]:
                    raise XimilarClientException(result["status"]["code"], result["status"]["text"])
                raise XimilarClientException(500, "Unexpected error when getting result from endpoint!")

    def resize_image_data(self, image_data, aspect_ratio=True, resize=True):
        """
        Resize image data that are no bigger than max_size.
        :param image_data: cv2/np ndarray
        :return: cv2/np ndarray
        """
        # do not resize image if set to 0
        if not resize or self.max_image_size == 0:
            return image_data

        import cv2

        height, width, _ = image_data.shape
        if height > self.max_image_size and width > self.max_image_size and not aspect_ratio:
            image_data = cv2.resize(image_data, (self.max_image_size, self.max_image_size))
        if height > self.max_image_size and width > self.max_image_size and aspect_ratio:
            image_data = cv2.resize(image_data, self.get_aspect_ratio_dim(image_data, self.max_image_size))
        return image_data

    def get_aspect_ratio_dim(self, image, img_size):
        if image.shape[0] > image.shape[1]:
            r = float(img_size) / image.shape[1]
            dim = (img_size, int(image.shape[0] * r))
        else:
            r = float(img_size) / image.shape[0]
            dim = (int(image.shape[1] * r), img_size)
        return dim

    def cv2_imread(self, path):
        import cv2

        image = cv2.imread(str(path))
        return image

    def cv2_imwrite(self, image_record, path):
        """
        Writes image data into given file
        :param image_record: dictionary with image data and color space
        :param path: file to store the image to
        :return: true if everything was fine
        """
        import cv2

        image = image_record[IMG_DATA]
        image_space = image_record[COLOR_SPACE] if COLOR_SPACE in image_record else "RGB"
        image = self._convert_image_to_bgr(image, image_space)

        cv2.imwrite(path, image)
        return True

    def load_base64_file(self, path, resize=True):
        """
        Load file from disk to base64.
        :param path: local path to the image
        :return: base64 encoded string
        """
        # if the image is quite small then just convert it to base64
        if (not resize) or os.stat(path).st_size / (1024 * 1024) < 0.1:
            with open(path, "rb") as image_file:
                encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
            return encoded_string

        # otherwise convert it to cv2 matrix, then encode it  and then to base64
        image = self.cv2_imread(path)
        image = self.cv2img_to_base64(image, image_space="BGR", resize=resize)
        return image

    def cv2img_to_base64(self, image, image_space="RGB", resize=True):
        """
        Load raw numpy/cv2 data of image to base64. The input image to this method should have RGB order.
        The ximilar accepts base64 data to have BGR order that is why we convert it here.
        The image_data was loaded in similar way:
            image = cv2.imread(str(path))
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        :param image: numpy/cv2 data with RGB order
        :param image_space: from which color space we are converting (default RGB)
        :param resize: if we want to resize image
        :return: base64 encoded string
        """
        import cv2

        image = self._convert_image_to_bgr(image, image_space)
        image = self.resize_image_data(image, resize=resize)
        retval, buffer = cv2.imencode(".jpg", image, params=[cv2.IMWRITE_JPEG_QUALITY, 99])
        jpg_as_text = "data:image/jpeg;base64," + base64.b64encode(buffer).decode("utf-8")
        return jpg_as_text

    def base64_to_cv2img(self, base64image):
        """
        Convert base64 data to image in BGR format.
        :param base64image: base64 image encoded
        :return: opencv2/numpy image
        """
        try:
            import cv2
            import numpy

            header = BASE64_HEADER_PATTERN.match(base64image)
            if header:
                base64data = BASE64_HEADER_PATTERN.sub("", base64image)
                base64image = base64data

            img_array = base64.b64decode(base64image)

            image = numpy.frombuffer(img_array, dtype=numpy.uint8)
            image = cv2.imdecode(image, 1)
            if image.shape[2] != 3:
                raise Exception("Image has not shape (height, width, 3)")
            return image
        except Exception as e:
            raise Exception("Unable to read base64:" + str(e))

    def load_url_image(self, path, resize=True, timeout=30):
        """
        Load url file to base64 WITHOUT resizing (it is used in upload image for recognition).
        :param path: url path
        :param resize: if we want to resize image
        :param timeout: request timeout in seconds
        :return: base64 encoded string
        """
        import cv2
        import numpy

        r = requests.get(str(path), headers={"Accept": "*/*", "User-Agent": "request"}, timeout=timeout)
        img_bin = r.content
        image = numpy.asarray(bytearray(img_bin), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = self.cv2img_to_base64(image, resize=resize)
        return image

    def preprocess_records(self, records):
        """
        Preprocess all records (list of dictionaries with possible '_base64'|'_file'|'_url' fields
        before processing/upload to Ximilar Application.
        :param records: list of dictionaries
        :return: modified list of dictionaries
        """
        # (shallow) copy the records in order not to modify the incoming dictionaries
        records = [rec.copy() for rec in records]
        for i in range(len(records)):
            if (
                FILE not in records[i]
                and BASE64 not in records[i]
                and URL not in records[i]
                and IMG_DATA not in records[i]
                and "_text_data" not in records[i]
                and "_id" not in records[i]
            ):
                raise Exception("Please specify one of '_file', '_base64', '_url', '_img_data' field in record")

            noresize = NORESIZE in records[i] and records[i][NORESIZE]

            if FILE in records[i] and BASE64 not in records[i] and IMG_DATA not in records[i]:
                records[i][BASE64] = self.load_base64_file(records[i][FILE], resize=not noresize)
            elif (BASE64 in records[i] or URL in records[i]) and (noresize or self.max_image_size == 0):
                # if we have base64 and we do not want to resize it at all
                pass
            elif BASE64 in records[i]:
                # if we have base64 and we need to resize it
                image = self.base64_to_cv2img(records[i][BASE64])
                records[i][BASE64] = self.cv2img_to_base64(image, image_space="BGR", resize=not noresize)
            elif IMG_DATA in records[i]:
                records[i][BASE64] = self.cv2img_to_base64(
                    records[i][IMG_DATA],
                    image_space=records[i][COLOR_SPACE] if COLOR_SPACE in records[i] else "RGB",
                    resize=not noresize,
                )

            # finally we need to delete the image data and just send url or base64
            if IMG_DATA in records[i]:
                del records[i][IMG_DATA]

        return records

    def custom_endpoint_processing(self, records, endpoint, request_params=None):
        """
        Records processing for your custom endpoint.
        :param records: list of dictionaries with _url, _file, _base64
        :param endpoint: endpoint to add to api.ximilar.com/
        :return: result from endpoint
        """
        data = {RECORDS: self.preprocess_records(records)}
        if request_params and isinstance(request_params, dict):
            data.update(request_params)
        return self.post(endpoint, data=data)

    def get_config(self, config_type, version=None):
        """
        Get configuration json from account/v2/config endpoint
        :param config_type: name of the configuration
        :param version: number of config version (optional
        :return: json response
        """
        version = "?version=" + str(version) if version else ""
        return self.get(CONFIG_ENDPOINT + config_type + version)

    @staticmethod
    def download_image(url, destination="", name=None):
        """
        Download image from url to the destination
        :param url: url to the image
        :param destination: where the image will be stored
        :return: None
        """
        f_name = url.split("/")[-1] if name is None else name
        f_name = f_name.split("?")[0]
        f_dest = os.path.join(destination, f_name)
        f_dest += ".jpg" if "." not in f_name else ""

        if os.path.isfile(f_dest):
            return f_dest

        page = requests.get(url)
        with open(f_dest, "wb") as f:
            f.write(page.content)
        return f_dest

    def parallel_records_processing(
        self, records, method, max_workers=3, batch_size=1, output=False, error_file="", timeout=60
    ):
        """
        Process method which uses records in parallel way. This works for methods:

        RecogntionClient.upload_image
        RecognitionClient.classify
        FashionTaggingClient.tags
        GenericTaggingClient.tags
        DominantColorClient.dominantcolor

        :param records: list of dictionaries with _url, _file, _base64 and other metadata
                        [{'_file': '__IMG_PATH__', 'id': '__MY_IMAGE_ID__'}, ... ]
        :param method: method to call
        :param max_workers: how many threads will we spawn for work (recommended is 3)
        :param batch_size: how many images are we sending in batch
        :param output: output to stdout with progressbar / tqdm
        :return: list of results from every method
        """

        def catch_exception_method(method_to_process, records):
            try:
                return method_to_process(records)
            except Exception as e:
                return None

        executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        status = {"answer_records": 0, "records": 0, "error": 0, "skipped_records": 0}
        errors, results = [], []

        with tqdm(total=len(records), disable=not output) as pbar:
            for brecords in self.batch(records, n=batch_size * max_workers * 2):
                futures = []
                for records_to_proc in self.batch(brecords, n=batch_size):
                    futures.append(
                        {
                            "future": executor.submit(catch_exception_method, method, records_to_proc),
                            "records": records_to_proc,
                            "size": len(records_to_proc),
                        }
                    )

                for future in futures:
                    try:
                        result = future["future"].result(timeout=timeout)
                        self.update_status(status, result, future["records"], errors)
                        if result is not None:
                            results.append(result)
                    except Exception as e:
                        print("Error processing records: ")
                        print(future["records"])
                        print("Exception:", str(e))
                    pbar.update(future["size"])

        if error_file and len(errors) > 0:
            print(f"Writting error log to file: {error_file}")
            with open(error_file, "w") as f:
                json.dump(errors, f, indent=4)

        if output:
            print("Result of parallel: ", status)

        return results

    def update_status(self, status, result, records, errors):
        if result is None:
            return

        if "status" in result and (
            "records" not in result and "skipped_records" not in result and "answer_records" not in result
        ):
            if isinstance(result["status"], int):
                if result["status"] == 500:
                    for record in records:
                        status["error"] += 1
                        errors.append(record)
                elif result["status"] >= 300:
                    status["error"] += 1
            if isinstance(result["status"], dict):
                if "code" in result["status"]:
                    if result["status"]["code"] >= 300:
                        status["error"] += 1
        if "records" in result:
            status["records"] += len(result["records"])
        if "skipped_records" in result:
            for skipped in result["skipped_records"]:
                if "_reason" in skipped and skipped["_reason"]["status"]["code"] != 211:
                    errors.append(skipped)
                    status["error"] += 1
                else:
                    status["skipped_records"] += 1
        if "answer_records" in result:
            status["answer_records"] += len(result["answer_records"])

    def batch(self, iterable, n=1):
        """
        Divides list to batches of size n.
        """
        l = len(iterable)
        for ndx in range(0, l, n):
            yield iterable[ndx : min(ndx + n, l)]

    @staticmethod
    def _convert_image_to_bgr(image, image_space="RGB"):
        """
        Converts given image from given record to BGR
        :param image: image data
        :param image_space: string saying the image space
        :return: image data in BGR
        """
        import cv2

        if image_space == "BGR":
            return image
        if image_space == "RGB":
            return cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        elif image_space == "HSV":
            return cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
        elif image_space == "LUV":
            return cv2.cvtColor(image, cv2.COLOR_LUV2BGR)
        raise ValueError("unknown image space: " + image_space)



from .client import RestClient
from .recognition import RecognitionClient
from .tagging import GenericTaggingClient, FashionTaggingClient, HomeDecorTaggingClient
from .colors import DominantColorProductClient, DominantColorGenericClient
from .detection import DetectionClient
from .search import (
    SimilarityPhotosClient,
    SimilarityProductsClient,
    SimilarityFashionClient,
    SimilarityCustomClient,
    ImageMatchingSearchClient,
    SimilarityColorsClient,
    SimilarityHomeDecorClient,
)
from .collectibles import CardGradingClient, CollectiblesRecognitionClient, CardConditionClient, CardCenteringClient
from .flows import FlowsClient
from .exceptions import XimilarClientException
from .removebg import RemoveBGClient
from .similarity import CustomSimilarityClient
from .upscaler import UpscaleClient
from .asyncr import AsyncRClient, AsynchronousRequest
from .ocr import OCRClient
from .description import ProductDescriptionClient



# ----------------- PLEASE READ -------------------
# RUN THIS as pytest client.py --token __TOKEN_ID__
# This test scenario suggest that you already:
#     1. In your Default workspace you have on Tagging and one Categorization task trained
#     2. Also that you have resources to almost all services
#     3. In your Default workspace some images are present
#     4. You have valid flow

import pytest

from ximilar.client.constants import *
from ximilar.client.recognition import RecognitionClient, Image, Label, Task
from ximilar.client.detection import DetectionClient, DetectionObject, DetectionLabel, DetectionTask
from ximilar.client.tagging import FashionTaggingClient, GenericTaggingClient
from ximilar.client.colors import DominantColorProductClient, DominantColorGenericClient
from ximilar.client.flows import FlowsClient

TASK_NAME = "Test-Task-In-Vize-X-1"
LABEL_NAME = "Test-Task-In-Vize-Label-X-1"
TEST_IMG_URL = "https://vize.ai/examples/screw/05.jpg"
TEST_IMG_BIG = "big_image.jpg"
TEST_IMG_SMALL = "ximilar.png"


def get_recognition_client(request) -> RecognitionClient:
    token = request.config.getoption("--token")
    client = RecognitionClient(token)
    return client


def get_detection_client(request):
    token = request.config.getoption("--token")
    client = DetectionClient(token)
    return client


def test_01_client_non_existing(request):
    """
    Tests an API call for non existing Token
    """
    with pytest.raises(Exception) as e:
        token = "non-existing-token"
        client = RecognitionClient(token)

    assert "tech@ximilar.com" in str(e.value)


def test_02_client_existing(request):
    """
    Tests an API call for existing Token
    """
    client = get_recognition_client(request)
    tasks, status = client.get_all_tasks()

    assert isinstance(tasks, list)
    if len(tasks):
        assert isinstance(tasks[0], Task)
        assert hasattr(tasks[0], ID)
        assert hasattr(tasks[0], NAME)
        assert hasattr(tasks[0], TYPE)
        assert hasattr(tasks[0], "production_version")


def test_03_client_all_labels(request):
    """
    Tests an API call for all labels
    """
    client = get_recognition_client(request)
    labels, status = client.get_all_labels()

    assert isinstance(labels, list)
    if len(labels):
        assert isinstance(labels[0], Label)
        assert hasattr(labels[0], ID)
        assert hasattr(labels[0], NAME)


def test_04_client_create_task_label_image(request):
    """
    Tests an API call for creating task, label, image.
    """
    # Query to test
    client = get_recognition_client(request)
    labels_before, status = client.get_all_labels()

    # LETS create some task, label and image
    task, status = client.create_task(TASK_NAME)
    label, status = task.create_label(LABEL_NAME)
    images0, n_page, status = label.get_training_images()
    images, status = label.upload_images([{FILE: TEST_IMG_SMALL}])
    images1, n_page, status = label.get_training_images()
    added_labels, status = images[0].get_labels()
    cached_labels, status = images[0].get_labels()

    # Query after creating
    task1, status = client.get_task(task.id)
    labels1, status = task1.get_labels()
    label.detach_image(images[0].id)
    tasksX, status = client.get_tasks_by_name(TASK_NAME)
    labelsX, status = client.get_labels_by_substring(LABEL_NAME)

    # then delete everything
    client.remove_image(images[0].id)
    images2, n_page, status = label.get_training_images()
    task.detach_label(label.id)

    for task_to_delete in tasksX:
        task_to_delete.remove()
    for label_to_delete in labelsX:
        label_to_delete.remove()

    # query after deleting
    tasks, status = client.get_tasks_by_name(TASK_NAME)
    labels, status = client.get_labels_by_substring(LABEL_NAME)
    labels_after, status = client.get_all_labels()
    no_task, no_task_status = client.get_task(task.id)
    no_image, no_image_status = client.get_label(label.id)
    no_label, no_label_status = client.get_label(label.id)

    # check the label had connected images
    assert len(added_labels) == 1
    assert len(cached_labels) == 1

    # and check it
    assert labels is None
    assert tasks is None
    assert no_task is None
    assert no_image is None
    assert no_label is None

    # check creating of Task, Label and Image
    assert isinstance(task, Task)
    assert isinstance(label, Label)
    assert isinstance(images[0], Image)

    # check the get endpoints
    assert task.id == tasksX[0].id
    assert label.id == labelsX[0].id

    # check the creating of label and image
    assert len(labels_before) == len(labels_after)
    assert len(labelsX) == 1
    assert len(images0) == 0
    assert len(images1) == 1
    assert len(images2) == 0


def test_05_paginated_methods_recognition_1(request):
    """
    Test creating labels and getting all labels(which works with pagination.
    """
    client = get_recognition_client(request)
    labels_old, status = client.get_all_labels()

    labels = []
    for i in range(10):
        label, status = client.create_label(LABEL_NAME)
        labels.append(label)

    all_labels, status = client.get_all_labels()

    for label in labels:
        label.remove()

    assert len(all_labels) == len(labels_old) + len(labels)


def test_06_upload_image_url_file_big(request):
    """
    Test uploading images with resize and no resize
    """
    client = get_recognition_client(request)
    images1, status = client.upload_images([{URL: TEST_IMG_URL, "noresize": True}])
    images2, status = client.upload_images([{URL: TEST_IMG_URL}])
    images3 = client.parallel_records_processing([{FILE: TEST_IMG_BIG, "noresize": True}], client.upload_images)

    images1[0].remove()
    images2[0].remove()
    images3[0][0][0].remove()

    assert images2[0].img_width == client.max_image_size
    assert images2[0].img_height >= client.max_image_size
    assert images1[0].img_height > images2[0].img_height
    assert images1[0].img_width > images2[0].img_width
    assert images3[0][0][0].img_height == images1[0].img_height
    assert images3[0][0][0].img_width == images1[0].img_width
    assert 0 == 0


def test_07_upload_image_url_file_small(request):
    """
    Test uploading images with resize and no resize
    """
    client = get_recognition_client(request)
    images1, status = client.upload_images([{FILE: TEST_IMG_SMALL, "noresize": True}])
    images2, status = client.upload_images([{FILE: TEST_IMG_SMALL}])

    images1[0].remove()
    images2[0].remove()

    assert images2[0].img_height == images1[0].img_height


def test_08_classify(request):
    """
    Tests an API call for one of our label
    """
    client = get_recognition_client(request)
    tasks, status = client.get_all_tasks()
    task = tasks[0]

    result = task.classify([{FILE: TEST_IMG_SMALL}])
    assert isinstance(result, dict)


def test_09_parallel_classify(request):
    """
    Test an parallel method for classification endpoint.
    """
    client = get_recognition_client(request)
    tasks, status = client.get_all_tasks()
    task = tasks[0]

    result = client.parallel_records_processing(
        [{FILE: TEST_IMG_SMALL} for i in range(3)], method=task.classify, output=False, max_workers=3
    )
    assert len(result) == 3 and RECORDS in result[0] and len(result[0][RECORDS]) == 1


def test_10_fashion_tagging(request):
    """
    Test fashion tagging system prediction.
    """
    client = FashionTaggingClient(request.config.getoption("--token"))
    result = client.tags([{FILE: TEST_IMG_SMALL}])

    assert RECORDS in result and len(result[RECORDS]) > 0
    assert "_tags" in result[RECORDS][0]


def test_11_parallel_fashion_processing(request):
    """
    Test parallel processing for fashion tagging prediction.
    """
    client = FashionTaggingClient(request.config.getoption("--token"))
    result3 = client.parallel_records_processing(
        [{FILE: TEST_IMG_SMALL} for i in range(3)], method=client.tags, output=False, max_workers=3
    )
    result1 = client.parallel_records_processing(
        [{FILE: TEST_IMG_SMALL} for i in range(3)], method=client.tags, output=False, max_workers=3, batch_size=3
    )

    assert len(result1) == 1 and RECORDS in result1[0] and len(result1[0][RECORDS]) == 3
    assert len(result3) == 3 and RECORDS in result3[0] and len(result3[0][RECORDS]) == 1


def test_12_generic_tagging(request):
    """
    Test generic tagging system for prediction.
    """
    client = GenericTaggingClient(request.config.getoption("--token"))
    result = client.tags([{FILE: TEST_IMG_SMALL}])

    assert RECORDS in result and len(result[RECORDS]) > 0
    assert "_tags" in result[RECORDS][0]


def test_13_detection_workflow(request):
    """
    This is simple workflow test for detection.
    """
    client = get_detection_client(request)

    # lets create task, labels, image
    task, status = client.create_task("Test-Detection-Task-API")
    label1, status = client.create_label("Test-Detection-Label-API-1")
    label2, status = client.create_label("Test-Detection-Label-API-2")
    images1, status = client.upload_images([{FILE: TEST_IMG_BIG, "noresize": True}])
    object1, status = client.create_object(label1.id, images1[0].id, [0, 0, 100, 100])

    # connect label to task
    task.add_label(label1.id)

    # get labels of task
    labels1, status = task.get_labels()

    # get all detection labels
    labels2, status = client.get_all_labels()
    objects, status = client.get_objects_of_image(images1[0].id)

    # remove everything
    task.remove()
    label1.remove()
    label2.remove()
    images1[0].remove()  # this should delete also object
    object1.remove()

    assert len(objects) == 1
    assert len(labels1) == 1
    assert len(labels1) != len(labels2)


def test_14_upload_and_verify_image(request):
    """
    Test uploading images, verify and unverify them.
    """
    client = get_recognition_client(request)
    user = client.get_user_details()

    images1, status = client.upload_images([{URL: TEST_IMG_URL}])

    images1[0].verify(user["id"])
    verify_cnt = images1[0].verifyCount
    images1[0].unverify()
    unverify_cnt = images1[0].verifyCount

    images1[0].remove()

    assert verify_cnt == 1
    assert unverify_cnt == 0


def test_15_dominant_color_product(request):
    """
    Test dominant color product system for prediction.
    """
    client = DominantColorProductClient(request.config.getoption("--token"))
    result = client.dominantcolor([{FILE: TEST_IMG_SMALL}])

    assert RECORDS in result and len(result[RECORDS]) > 0
    assert "_dominant_colors" in result[RECORDS][0]


def test_16_dominant_color_generic(request):
    """
    Test dominant color generic system for prediction.
    """
    client = DominantColorGenericClient(request.config.getoption("--token"))
    result = client.dominantcolor([{FILE: TEST_IMG_SMALL}])

    assert RECORDS in result and len(result[RECORDS]) > 0
    assert "_dominant_colors" in result[RECORDS][0]


def test_17_test_flow(request):
    """
    Test flow for prediction.
    """
    client = FlowsClient(request.config.getoption("--token"))
    flows, _ = client.get_all_flows()

    for flow in flows:
        if flow.valid:
            result = flow.process([{FILE: TEST_IMG_SMALL}])
            json_flow = flow.to_json()

            assert RECORDS in result and len(result[RECORDS]) > 0
            assert "type" in json_flow and "id" in json_flow
            break



import requests

from ximilar.client import RestClient
from ximilar.client.constants import *

FASHION_TAGGING_ENDPOINT = "tagging/fashion/v2/tags"
META_FASHION_TAGGING_ENDPOINT = "tagging/fashion/v2/meta"
GENERIC_TAGGING_ENDPOINT = "photo/tags/v2/tags"
HOME_DECOR_TAGGING_ENDPOINT = "tagging/homedecor/v2/tags"
DETECT_HOME_DECOR_TAGGING_ENDPOINT = "tagging/homedecor/v2/detect_tags"
DETECT_FASHION_TAGGING_ENDPOINT = "tagging/fashion/v2/detect_tags"
DETECT_FASHION_ENDPOINT = "tagging/fashion/v2/detect"
DETECT_FASHION_TAGGING_ALL_ENDPOINT = "tagging/fashion/v2/detect_tags_all"
COLLECTIBLES_RECOGNITION_ENDPOINT = "collectibles/v2/process"
COLLECTIBLES_ANALYZE_ENDPOINT = "collectibles/v2/analyze"


class TaggingClient(RestClient):
    def construct_data(self, records, aggregate_labels=None, profile=None, **kwargs):
        if len(records) == 0:
            raise Exception("Please specify at least on record when using tags method.")

        data = {RECORDS: self.preprocess_records(records)}

        if aggregate_labels:
            data["aggregate_labels"] = True

        if profile:
            data["profile"] = profile

        if kwargs:
            data.update(kwargs)

        return data

    def tags(self, records, endpoint, aggregate_labels=False, profile=None, **kwargs):
        """
        Call the tagging endpoint
        :param records: [description]
        :param endpoint: [description]
        :param aggregate_labels: [description], defaults to False
        :param profile: [description], defaults to None
        :return: json result data from the API
        """
        data = self.construct_data(records, aggregate_labels=aggregate_labels, profile=profile, **kwargs)
        result = self.post(endpoint, data=data)
        # todo: self.check_json_status(result)
        return result


class FashionTaggingClient(TaggingClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=FASHION_TAGGING):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = FASHION_TAGGING_ENDPOINT

    def change_predict_endpoint(self, new_endpoint):
        self.PREDICT_ENDPOINT = new_endpoint

    def tags(self, records, aggregate_labels=False, profile=None):
        return super().tags(records, self.PREDICT_ENDPOINT, aggregate_labels=aggregate_labels, profile=profile)

    def meta_tags(self, records, profile=None, **kwargs):
        return super().tags(records, META_FASHION_TAGGING_ENDPOINT, profile=profile, **kwargs)

    def detect(self, records, profile=None, **kwargs):
        return super().tags(records, DETECT_FASHION_ENDPOINT, profile=profile, **kwargs)

    def detect_tags(self, records, profile=None, **kwargs):
        return super().tags(records, DETECT_FASHION_TAGGING_ENDPOINT, profile=profile, **kwargs)

    def detect_tags_all(self, records, profile=None, **kwargs):
        return super().tags(records, DETECT_FASHION_TAGGING_ALL_ENDPOINT, profile=profile, **kwargs)

    def get_top_categories(self):
        result = requests.get(self.urljoin(self.endpoint, "tagging/fashion/v2/top_categories"))
        return result.json()["labels"]

    def get_categories(self):
        result = requests.get(self.urljoin(self.endpoint, "tagging/fashion/v2/categories"))
        return result.json()["labels"]


class HomeDecorTaggingClient(TaggingClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=HOME_DECOR_TAGGING):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = HOME_DECOR_TAGGING_ENDPOINT

    def change_predict_endpoint(self, new_endpoint):
        self.PREDICT_ENDPOINT = new_endpoint

    def tags(self, records, aggregate_labels=False, profile=None):
        return super().tags(records, self.PREDICT_ENDPOINT, aggregate_labels=aggregate_labels, profile=profile)

    def detect_tags(self, records, profile=None, **kwargs):
        return super().tags(records, DETECT_HOME_DECOR_TAGGING_ENDPOINT, profile=profile, **kwargs)


class GenericTaggingClient(TaggingClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=GENERIC_TAGGING):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = GENERIC_TAGGING_ENDPOINT

    def tags(self, records):
        return super().tags(records, self.PREDICT_ENDPOINT)



from ximilar.client import RestClient
from ximilar.client.constants import *

PRODUCT_ENDPOINT = "product-description/v2/generate"


class ProductDescriptionClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="product-description"):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = PRODUCT_ENDPOINT

    def construct_data(self, records=[], style_name=None, product_info=None, product_type=None, tagging_type=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")
        data = {RECORDS: self.preprocess_records(records)}

        if style_name:
            data["style_name"] = style_name
        if product_info:
            data["product_info"] = product_info
        if product_type:
            data["product_type"] = product_type
        if tagging_type:
            data["tagging_type"] = tagging_type

        return data

    def generate(self, records, style_name=None, product_info=None, product_type=None, tagging_type=None):
        records = self.preprocess_records(records)
        return self.post(
            PRODUCT_ENDPOINT, data=self.construct_data(records, style_name, product_info, product_type, tagging_type)
        )



from ximilar.client import RestClient
from ximilar.client.constants import *

GRADING_ENDPOINT = "card-grader/v2/grade"
CONDITION_ENDPOINT = "card-grader/v2/condition"
CENTERING_ENDPOINT = "card-grader/v2/centering"

COLLECTIBLES_PROCESS = "collectibles/v2/process"
COLLECTIBLES_COMICS_ID = "collectibles/v2/comics_id"
COLLECTIBLES_CARD_ID = "collectibles/v2/tcg_id"
COLLECTIBLES_SPORT_ID = "collectibles/v2/sport_id"
COLLECTIBLES_DETECT = "collectibles/v2/detect"
COLLECTIBLES_SLAB_ID = "collectibles/v2/slab_id"
COLLECTIBLES_OCR_ID = "collectibles/v2/ocr_id"
COLLECTIBLES_ANALYZE = "collectibles/v2/analyze"


class CardGradingClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="card-grader"):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = GRADING_ENDPOINT

    def construct_data(self, records=[]):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")
        data = {RECORDS: self.preprocess_records(records)}
        return data

    def grade(self, records, endpoint=GRADING_ENDPOINT):
        records = self.preprocess_records(records)
        return self.post(endpoint, data={RECORDS: records})


class CardConditionClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="card-grader"):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = CONDITION_ENDPOINT

    def construct_data(self, records=[], mode=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")
        data = {RECORDS: self.preprocess_records(records)}
        if mode:
            data["mode"] = mode
        return data

    def grade(self, records, mode=None, endpoint=CONDITION_ENDPOINT):
        return self.post(endpoint, data=self.construct_data(records, mode))


class CardCenteringClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="card-grader"):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = CENTERING_ENDPOINT

    def construct_data(self, records=[], mode=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")
        data = {RECORDS: self.preprocess_records(records)}
        return data

    def grade(self, records, mode=None, endpoint=CENTERING_ENDPOINT):
        return self.post(endpoint, data=self.construct_data(records, mode))


class CollectiblesRecognitionClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=COLLECTIBLES_RECOGNITION):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = COLLECTIBLES_ANALYZE

    def construct_data(self, records=[], lang=False, slab_id=False, slab_grade=False, fields_to_return=None, analyze_all=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")
        data = {RECORDS: self.preprocess_records(records)}

        if lang:
            data["lang"] = lang

        if slab_id:
            data["slab_id"] = slab_id

        if fields_to_return:
            data["fields_to_return"] = fields_to_return

        if slab_grade:
            data["slab_grade"] = slab_grade

        if analyze_all:
            data["analyze_all"] = analyze_all

        return data

    def process(self, records: list):
        data = {RECORDS: self.preprocess_records(records)}
        result = self.post(COLLECTIBLES_PROCESS, data=data)
        return result

    def card_id(
        self, records: list, lang: bool = False, slab_id: bool = False, slab_grade: bool = False, fields_to_return=None, analyze_all=None
    ):
        result = self.post(
            COLLECTIBLES_CARD_ID,
            data=self.construct_data(
                records, lang=lang, slab_id=slab_id, slab_grade=slab_grade, fields_to_return=fields_to_return, analyze_all=analyze_all
            ),
        )
        return result

    def sport_id(self, records: list, slab_id: bool = False, slab_grade: bool = False, fields_to_return=None):
        result = self.post(
            COLLECTIBLES_SPORT_ID,
            data=self.construct_data(
                records, slab_id=slab_id, slab_grade=slab_grade, fields_to_return=fields_to_return
            ),
        )
        return result

    def comics_id(self, records: list, slab_id: bool = False, slab_grade: bool = False, fields_to_return=None):
        result = self.post(
            COLLECTIBLES_COMICS_ID,
            data=self.construct_data(
                records, slab_id=slab_id, slab_grade=slab_grade, fields_to_return=fields_to_return
            ),
        )
        return result

    def detect(self, records: list):
        data = {RECORDS: self.preprocess_records(records)}
        result = self.post(COLLECTIBLES_DETECT, data=data)
        return result

    def slab_id(self, records: list):
        data = {RECORDS: self.preprocess_records(records)}
        result = self.post(COLLECTIBLES_SLAB_ID, data=data)
        return result

    def ocr_id(self, records: list):
        data = {RECORDS: self.preprocess_records(records)}
        result = self.post(COLLECTIBLES_OCR_ID, data=data)
        return result

    def analyze(self, records: list, fields_to_return=None):
        result = self.post(COLLECTIBLES_ANALYZE, data=self.construct_data(records, fields_to_return=fields_to_return))
        return result



from ximilar.client import RestClient
from ximilar.client.constants import *

FLOW_ENDPOINT = "/flows/v2/flow/"


class FlowsClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="flows", max_image_size=512):
        super().__init__(token=token, endpoint=endpoint, max_image_size=max_image_size, resource_name=resource_name)

        self.PREDICT_ENDPOINT = "/flows/v2/process"

    def construct_data(self, flow=None, records=None, store_images=None):
        if flow is None or records is None or len(records) == 0:
            raise Exception("Please specify flow and records!")

        data = {"flow": flow, RECORDS: self.preprocess_records(records)}
        if store_images is not None:
            data[STORE_IMAGES] = store_images

        return data

    def get_all_flows(self):
        flows, status = self.get_all_paginated_items(FLOW_ENDPOINT)

        if not flows and status[STATUS] == STATUS_ERROR:
            return None, status

        return [Flow(self.token, self.endpoint, flow_json) for flow_json in flows], RESULT_OK

    def get_flow(self, flow_id):
        flow_json = self.get(FLOW_ENDPOINT + flow_id)

        if DETAIL in flow_json:
            return None, RESULT_ERROR

        return Flow(self.token, self.endpoint, flow_json, max_image_size=self.max_image_size), RESULT_OK

    def get_flow_json(self, flow_id):
        return self.get(FLOW_ENDPOINT + str(flow_id) + "/json")

    def process_flow(self, flow, records, store_images=None, headers=None):
        data = self.construct_data(flow=flow, records=records, store_images=store_images)
        return self.post(self.PREDICT_ENDPOINT, data=data, headers=headers)


class Flow(FlowsClient):
    """
    Flow entity from /flows/v2/flow endpoint.
    """

    def __init__(self, token, endpoint, flow_json, max_image_size=512):
        super().__init__(token, endpoint=endpoint, max_image_size=max_image_size, resource_name="")

        self.id = flow_json[ID]
        self.name = flow_json[NAME]
        self.description = flow_json[DESCRIPTION]
        self.workspace = flow_json[WORKSPACE]
        self.top_node = flow_json[TOP_NODE]
        self.valid = flow_json[VALID]
        self.image_size = int(flow_json["image_size"]) if "image_size" in flow_json else 512

    def to_json(self):
        """
        Returns flow in json format.
        """
        return self.get(FLOW_ENDPOINT + str(self.id) + "/json")

    def process(self, records, store_images=None):
        """
        Call processing of the flow.
        """
        return self.process_flow(self.id, records, store_images=store_images)



from functools import wraps
import time


def retry_when(*exceptions, attempts=4, start_pause=5, multiply_pause=4, verbose=1):
    """
    Decorator which repeats given function in case any given exception occurs.
    By default, execute 4 times with 5, 20 and 80 second breaks between attempts.
    Exception is re-thrown only if number of attempts is exceeded.

    :param exceptions: one or more exceptions this decorator will catch
    :param attempts: how many times we will try to call the function; default 4
    :param start_pause: how many seconds we will be between first and second attempt; default 5
    :param multiply_pause: how the time between attempts will change; default 4 (pause is 4x longer after each failure)
    :param verbose: 0 for no prints, 1 print message when function call is no successful; default: 1
    """

    def decorator(function):
        @wraps(function)
        def wrap(*args, **kwargs):
            next_try_sec = start_pause

            for i in range(5):
                try:
                    return function(*args, **kwargs)
                except exceptions as e:
                    if verbose >= 1:
                        print(
                            f"{type(e).__name__} occurred. Attempt {i + 1}/{attempts}, "
                            + f"next try in {next_try_sec:.1f} seconds."
                        )
                    if i + 1 == attempts:
                        raise e

                    time.sleep(next_try_sec)
                    next_try_sec *= multiply_pause

        return wrap

    return decorator



from ximilar.client.recognition import RecognitionClient
from ximilar.client.constants import *


class CustomPredictClient(RecognitionClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name="custom-model"):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = "custom/v2/predict"

    def predict(self, records, task_id, version=None):
        """
        Takes the images and calls the ximilar client for extracting custom outputs.
        """
        # version is default set to None, so ximilar will determine which one to take
        data = self.construct_data(records=records, task_id=task_id, version=version)
        result = self.post(self.PREDICT_ENDPOINT, data=data)

        self.check_json_status(result)
        return result



import json

from ximilar.client import RestClient
from ximilar.client.constants import *

DOMINANT_COLOR_GENERIC_ENDPOINT = "dom_colors/generic/v2/dominantcolor"
DOMINANT_COLOR_PRODUCT_ENDPOINT = "dom_colors/product/v2/dominantcolor"


class DominantColorClient(RestClient):
    def construct_data(self, records=[]):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")

        data = {RECORDS: self.preprocess_records(records)}
        return data

    def dominantcolor(self, records, endpoint):
        records = self.preprocess_records(records)
        return self.post(endpoint, data={RECORDS: records})


class DominantColorGenericClient(DominantColorClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=DOMINANT_COLORS_PRODUCT):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = DOMINANT_COLOR_GENERIC_ENDPOINT

    def dominantcolor(self, records):
        return super().dominantcolor(records, DOMINANT_COLOR_GENERIC_ENDPOINT)


class DominantColorProductClient(DominantColorClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=DOMINANT_COLORS_GENERIC):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = DOMINANT_COLOR_PRODUCT_ENDPOINT

    def dominantcolor(self, records):
        return super().dominantcolor(records, DOMINANT_COLOR_PRODUCT_ENDPOINT)



import requests

from ximilar.client import RestClient
from ximilar.client.constants import *

RECORDS = "records"


class UpscaleClient(RestClient):
    def construct_data(self, records, **kwargs):
        if len(records) == 0:
            raise Exception("Please specify at least one record when using tags method.")

        data = {RECORDS: self.preprocess_records(records)}

        if kwargs:
            data.update(kwargs)

        return data

    def upscale(self, records, scale, model_type="gan", **kwargs):
        """
        Call the upscaling endpoint
        :param records: [description]
        :param endpoint: [description]
        :return: json result data from the API
        """
        data = self.construct_data(records, model_type=model_type, **kwargs)
        result = self.post(f"/upscaler/{scale}x/upscale", data=data)
        return result



class XimilarClientException(Exception):
    def __init__(self, code, msg=None):
        Exception.__init__(self, msg)

        self.code = code
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class XimilarClientInvalidDataException(XimilarClientException):
    """
    Raised when invalid input data occur
    """

    def __init__(self, msg=None):
        super().__init__(400, msg or "Invalid data")



# HTTP CODES
HTTP_NO_CONTENT_204 = 204
HTTP_UNAVAILABLE_503 = 503

# General
RECORDS = "records"
FILE = "_file"
URL = "_url"
BASE64 = "_base64"
IMG_DATA = "_img_data"
COLOR_SPACE = "_color_space"
RESULT_OK = {"status": "OK"}
RESULT_ERROR = {"status": "ERROR", "detail": "Unexpected error."}
ENDPOINT = "https://api.ximilar.com/"

# Services
CUSTOM_IMAGE_RECOGNITION = "custom-recognition"
CUSTOM_OBJECT_DETECTION = "custom-object-detection"
PHOTO_SIMILARITY = "photo-similarity"
PRODUCT_SIMILARITY = "product-similarity"
FASHION_SIMILARITY = "fashion-similarity"
FASHION_SIMPLE = "fashion-search-simple"
HOMEDECOR_SIMILARITY = "homedecor-similarity"
GENERIC_TAGGING = "generic-tagging"
FASHION_TAGGING = "fashion-tagging"
VISUAL_SEARCH = "visual-search"
DOMINANT_COLORS_PRODUCT = "dominant-colors"
DOMINANT_COLORS_GENERIC = "dominant-colors"
HOME_DECOR_TAGGING = "homedecor-tagging"
CUSTOM_SIMILARITY = "custom-similarity"
COLOR_SIMILARITY = "color-similarity"
REMOVEBG = "remove-bg"
IMAGE_MATCHING = "image-matching"
PRODUCTS_MANAGEMENT = "products-management"
COLLECTIBLES_RECOGNITION = "collectibles-recognition"

# Recognition/Vize related
USER_ID = "user_id"
USER = "user"
IMAGE_ID = "image_id"
CREATED = "created"
TYPE_ID = "type_id"
STATUS = "status"
STATUS_OK = "OK"
STATUS_ERROR = "ERROR"
RESULTS = "results"
ID = "id"
NAME = "name"
TASK = "task"
TASKS_COUNT = "tasks_count"
NEGATIVE_FOR_TASK = "negative_for_task"
TYPE = "type"
TASK_TYPE = "type"
LABEL_TYPE = "type"
MULTI_LABEL = "multi_label"
MULTI_CLASS = "multi_class"
REGRESSION = "regression"
DETECTION = "detection"
TAG = "tag"
CATEGORY = "category"
VALUE = "value"
WORKSPACE = "workspace"
IMAGE = "image"
DEFAULT_WORKSPACE = "Default"
IMAGES_COUNT = "images_count"
NORESIZE = "noresize"
LABELS = "labels"
LABEL_NAMES = "label_names"
DETAIL = "detail"
VERSION = "version"
LABEL_ID = "label_id"
TASK_ID = "task_id"
IMG_PATH = "img_path"
THUMB_IMG_PATH = "thumb_img_path"
IMG_WIDTH = "img_width"
IMG_HEIGHT = "img_height"
VERIFY_COUNT = "verifyCount"
TRAIN_STATUS = "train_status"
TASK_NAME = "task_name"
PRODUCTION_VERSION = "production_version"
NEXT = "next"
DATA = "data"
DETECTION_LABEL = "detection_label"
RECOGNITION_LABELS = "recognition_labels"
META_DATA = "meta_data"
DESCRIPTION = "description"
OWNER = "owner"
RECOGNITION_TASKS = "recognition_tasks"
LAST_TRAIN_STATUS = "last_train_status"
OBJECTS = "objects"
BEST_LABEL = "best_label"
TEST_IMAGE = "test_image"
STORE_IMAGES = "store_images"
OUTPUT_NAME = "output_name"
COLOR = "color"
REAL_IMAGE = "real"
VALUES = "values"

# Flows
TOP_NODE = "top_node"
VALID = "valid"

# Similarity search
FIELDS_TO_RETURN = "fields_to_return"
_ID = "_id"
FILTER = "filter"
COUNT = "count"
K_COUNT = "k"
QUERY_RECORD = "query_record"
COLLECTION = "collection"
COLLECTION_ID = "collection-id"
ANSWER_RECORDS = "answer_records"



from setuptools import setup, find_packages
import os
import re
import subprocess

with open("requirements.txt") as f:
    install_requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

version_re = re.compile(r"^v_(\d+\.\d+\.\d+)")


def get_version():
    """Update this version if you are releasing a new version of the client on PyPI."""
    return "1.37.0"


setup(
    name="ximilar-client",
    version=get_version(),
    description="The Ximilar App and Vize.ai Client.",
    url="https://gitlab.com/ximilar-public/ximilar-vize-api",
    author="Michal Lukac, David Novak and Ximilar.com Team",
    author_email="tech@ximilar.com",
    license="Apache 2.0",
    packages=find_packages(),
    keywords="machine learning, multimedia, json, rest, data",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    install_requires=install_requirements,
    include_package_data=True,
    zip_safe=False,
    namespace_packages=["ximilar"],
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from ximilar.client import RestClient
from ximilar.client.constants import *

FACE_DETECTION_ENDPOINT = "identity/v2/face"
PERSON_DETECTION_ENDPOINT = "identity/v2/person"


class XimilarIdentityClient(RestClient):
    def face(self, records):
        records = self.preprocess_records(records)
        return self.post(FACE_DETECTION_ENDPOINT, data={RECORDS: records})

    def person(self, records):
        records = self.preprocess_records(records)
        return self.post(PERSON_DETECTION_ENDPOINT, data={RECORDS: records})



from ximilar.client import RestClient
from ximilar.client.constants import *

REMOVEBG_ENDPOINT_PRECISE = "removebg/precise/removebg"
REMOVEBG_ENDPOINT_FAST = "removebg/fast/removebg"


class RemoveBGClient(RestClient):
    def __init__(self, token, endpoint=ENDPOINT, resource_name=REMOVEBG):
        super().__init__(token=token, endpoint=endpoint, resource_name=resource_name)
        self.PREDICT_ENDPOINT = REMOVEBG_ENDPOINT_PRECISE

    def construct_data(self, records=[]):
        if len(records) == 0:
            raise Exception("Please specify at least one record in detect method!")

        data = {RECORDS: self.preprocess_records(records)}
        return data

    def removebg_precise(self, records, image_format="png"):
        records = self.preprocess_records(records)
        return self.post(self.PREDICT_ENDPOINT, data={RECORDS: records, "image_format": image_format})

    def removebg_fast(self, records, image_format="png"):
        records = self.preprocess_records(records)
        return self.post(REMOVEBG_ENDPOINT_FAST, data={RECORDS: records, "image_format": image_format})



import requests
import json

from ximilar.client import RestClient
from ximilar.client.constants import *
from ximilar.client.exceptions import XimilarClientInvalidDataException

LABEL_ENDPOINT = "recognition/v2/label/"
TASK_ENDPOINT = "recognition/v2/task/"
MODEL_ENDPOINT = "recognition/v2/model/"
IMAGE_ENDPOINT = "recognition/v2/training-image/"
CLASSIFY_ENDPOINT = "recognition/v2/classify/"
POINT_ENDPOINT = "annotate/v2/point/"


class RecognitionClient(RestClient):
    """
    Base Client for Ximilar Custom Image Recognition Service.
    """

    def __init__(
        self,
        token,
        endpoint=ENDPOINT,
        workspace=DEFAULT_WORKSPACE,
        max_image_size=512,
        resource_name=CUSTOM_IMAGE_RECOGNITION,
    ):
        self.workspace = workspace  # this must be set before calling supers
        super().__init__(token=token, endpoint=endpoint, max_image_size=max_image_size, resource_name=resource_name)
        self.PREDICT_ENDPOINT = CLASSIFY_ENDPOINT

    def get(self, api_endpoint, data=None, params=None):
        """
        Calling get request to the API.
        """
        return super().get(api_endpoint, data=data, params=self.add_workspace(params, api_endpoint))

    def post(self, api_endpoint, data=None, files=None, params=None, method=requests.post):
        """
        Calling post request to the API.
        """
        return super().post(api_endpoint, data=self.add_workspace(data), files=files, params=params, method=method)

    def put(self, api_endpoint, data=None, files=None, params=None):
        """
        Calling put request to the API.
        """
        return super().put(api_endpoint, data=self.add_workspace(data), files=files, params=params)

    def delete(self, api_endpoint, data=None, params=None):
        """
        Calling delete request to the API.
        """
        return super().delete(api_endpoint, data=data, params=self.add_workspace(params))

    def get_task(self, task_id):
        """
        Getting recognition task by the id/uuid.
        """
        task_json = self.get(TASK_ENDPOINT + task_id)
        if ID not in task_json:
            status = {STATUS: task_json[DETAIL]} if DETAIL in task_json else {STATUS: "Not Found"}
            return None, status
        return Task(self.token, self.endpoint, task_json), RESULT_OK

    def get_label(self, label_id):
        """
        Getting recognition label by the id/uuid.
        """
        label_json = self.get(LABEL_ENDPOINT + label_id)
        if ID not in label_json:
            return None, {STATUS: "label with id not found"}
        return Label(self.token, self.endpoint, label_json), RESULT_OK

    def get_model(self, model_id):
        """
        Getting recognition model by the id/uuid.
        """
        model_json = self.get(MODEL_ENDPOINT + model_id)
        if ID not in model_json:
            status = {STATUS: model_json[DETAIL]} if DETAIL in model_json else {STATUS: "Not Found"}
            return None, status
        return Model(self.token, self.endpoint, model_json), RESULT_OK

    def get_image(self, image_id):
        """
        Getting recognition image by the id/uuid.
        """
        image_json = self.get(IMAGE_ENDPOINT + image_id)
        if ID not in image_json:
            return None, {STATUS: "image with id not found"}
        return Image(self.token, self.endpoint, image_json), RESULT_OK

    def get_all_tasks(self, suffix=""):
        """
        Get all tasks of the user(user is specified by client key).
        :return: List of Tasks
        """
        tasks, status = self.get_all_paginated_items(TASK_ENDPOINT + suffix)

        if not tasks and status[STATUS] == STATUS_ERROR:
            return None, status

        return [Task(self.token, self.endpoint, t_json) for t_json in tasks], RESULT_OK

    def get_all_labels(self, suffix=""):
        """
        Get all labels of the user(user is specified by client key).
        :return: List of labels
        """
        labels, status = self.get_all_paginated_items(LABEL_ENDPOINT + suffix)

        if not labels and status[STATUS] == STATUS_ERROR:
            return None, status

        for label in labels:
            label["workspace"] = self.workspace

        return [Label(self.token, self.endpoint, l_json) for l_json in labels], RESULT_OK

    def get_tasks_by_name(self, name):
        """
        Get all tasks with the name.
        """
        tasks, result = self.get_all_tasks()

        tasks_to_return = []
        if result[STATUS] == STATUS_OK:
            for task in tasks:
                if task.name == name:
                    tasks_to_return.append(task)
        else:
            return None, result

        if len(tasks_to_return) > 0:
            return tasks_to_return, result

        return None, {STATUS: "Task with this name not found!"}

    def add_label_to_image(self, image_id, label_id):
        """
        Add label to the image.
        """
        return self.post(IMAGE_ENDPOINT + image_id + "/add-label/", data={LABEL_ID: label_id})

    def get_training_images(self, page_url=None, verification=None, real=None, test=False):
        """
        Get paginated result of images from workspace.
        :param page_url: optional, select the specific page of images, default first page
        :param verification: optional, integer which says how many verifications should have the images
        :return: (list of images, next_page)
        """
        url = (
            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
            if page_url
            else IMAGE_ENDPOINT + "?page=1"
        )
        url += "&verified=" + str(verification) if verification is not None else ""

        if real is not None:
            url += "&real=" + str(real)

        if test:
            url += "&test=true"

        result = self.get(url)
        return (
            [Image(self.token, self.endpoint, image_json) for image_json in result[RESULTS]],
            result[NEXT],
            {"count": result["count"], STATUS: "ok"},
        )

    def training_images_iter(self, page_url=None, verification=None, batch_size=1, test=False):
        """
        Get iterator overall all of images from workspace.
        :param page_url: optional, can add new parameters or select different than first page
        :param verification: optional, integer which says how many verifications should have the images
        :return: iterator
        """
        images, next_page, status = self.get_training_images(page_url=page_url, verification=verification, test=test)
        while images:
            for image in self.batch(images, batch_size):
                if batch_size == 1:
                    yield image[0]
                else:
                    yield image
            if not next_page:
                break
            images, next_page, status = self.get_training_images(next_page)

    def modify_images(self, images=None, labels=None, add_labels=None, remove_labels=None, test=None, real=None):
        """
        Modifies images in batch. Enter either list of images OR labels to be applied as a filter.
        :param images: [description], defaults to None
        :param labels: [description], defaults to None
        :param add_labels: [description], defaults to None
        :param remove_labels: [description], defaults to None
        :param test: [description], defaults to None
        :param real: [description], defaults to None
        :raises XimilarClientInvalidDataException: [description]
        :return: [description]
        """
        if images is None:
            images = []
        if labels is None:
            labels = []
        if remove_labels is None:
            remove_labels = []
        if add_labels is None:
            add_labels = []

        if not (labels or images):
            raise XimilarClientInvalidDataException("Either images or labels must be specified")

        data = {"images": images, "labels": labels, "labels-add": add_labels, "labels-remove": remove_labels}

        if test is not None:
            if test:
                data["mark-test"] = True
            else:
                data["unmark-test"] = True

        if real is not None:
            if real:
                data["mark-real"] = True
            else:
                data["unmark-real"] = True

        data["workspace"] = self.workspace
        return self.post(api_endpoint=IMAGE_ENDPOINT + "update", data=data)

    def get_labels_by_substring(self, name):
        """
        Return all labels of the workspace which contains the substring.
        """
        labels, result = self.get_all_labels(suffix="?search=" + name)

        if len(labels) > 0:
            return labels, result
        else:
            return None, {STATUS: "No labels found"}

    def get_label_by_name(self, name, type=None):
        """
        Get label with specified name which also belongs to this task.
        """
        labels, result = self.get_labels_by_substring(name)
        if result[STATUS] == STATUS_OK:
            for label in labels:
                if label.name == name and (type is None or label.type == type):
                    return label, RESULT_OK
        else:
            return None, result

        return None, {STATUS: "Label with this name not found!"}

    def remove_model(self, model_id):
        """
        Remove recognition model by id/uuid.
        """
        return self.delete(MODEL_ENDPOINT + model_id + "/")

    def remove_task(self, task_id):
        """
        Remove recognition task by id/uuid.
        """
        return self.delete(TASK_ENDPOINT + task_id + "/")

    def remove_label(self, label_id):
        """
        Remove recognition label by id/uuid.
        """
        return self.delete(LABEL_ENDPOINT + label_id)

    def remove_image(self, image_id):
        """
        Remove recognition image by id/uuid.
        """
        return self.delete(IMAGE_ENDPOINT + image_id)

    def create_task(self, name, description=None, task_type=MULTI_CLASS):
        """
        Create task with given name.
        :param name: name of the task
        :param task_type: 'multi_class' (CATEGORIZATION default) or 'multi_label' (TAGGING)
        :return: Task object, status
        """
        data = {NAME: name, TASK_TYPE: task_type, DESCRIPTION: description}
        task_json = self.post(TASK_ENDPOINT, data=data)
        if ID not in task_json:
            msg = task_json[DETAIL] if DETAIL in task_json else "unexpected error"
            return None, {STATUS: msg}
        return Task(self.token, self.endpoint, task_json), RESULT_OK

    def create_label(self, name, description=None, output_name=None, label_type=CATEGORY, **kwargs):
        """
        Create label with given name.
        :param name: name of the label
        :param label_type: type of label to create (category or tag)
        :return: Label object, status
        """
        data = {NAME: name, LABEL_TYPE: label_type, DESCRIPTION: description, OUTPUT_NAME: output_name}
        data.update(kwargs)

        label_json = self.post(LABEL_ENDPOINT, data=data)
        if ID not in label_json:
            return None, {STATUS: "unexpected error"}
        return Label(self.token, self.endpoint, label_json), RESULT_OK

    def create_point(self, image_id=None, name=None, data=None):
        points_json = self.post(POINT_ENDPOINT, {"image": image_id, "name": name, "data": data})
        return points_json, STATUS_OK

    def get_points(self, image_id=None, name=None):
        """
        Get points stored (optional for image or filtered by name).
        """
        url = POINT_ENDPOINT
        if image_id:
            url += "?image=" + str(image_id)

        points_json = self.get(url)
        return [Point(self.token, self.endpoint, point) for point in points_json[RESULTS]], STATUS_OK

    def parse_already_inserted(self, detail):
        import re

        worst_status = {STATUS: "exists"}
        result = re.search(
            r".*image.ID..([a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89aAbB][a-f0-9]{3}-[a-f0-9]{12}).*", detail
        ).group(1)
        image, _ = self.get_image(result)
        return image, worst_status

    def upload_images(self, records):
        """
        Upload one or more files and add labels to them.
        :param records: list of dictionaries with labels and one of '_base64', '_file', '_url'
                        specify noresize: True to save image without (default False)
                       [{'_file': '__FILE_PATH__', 'labels': ['__UUID_1__', '__UUID_2__'], noresize: False}, ...]
        :return: image, status
        """
        images = []
        worst_status = RESULT_OK
        for record in records:
            files, data = None, None
            noresize = NORESIZE in record and record[NORESIZE]
            noresize_on_server = noresize or self.max_image_size > 1024
            metadata = record[META_DATA] if META_DATA in record and record[META_DATA] else {}
            test_image = record[TEST_IMAGE] if TEST_IMAGE in record else False

            data = self._create_image_data(record, noresize, noresize_on_server, test_image, metadata)

            image_json = self.post(IMAGE_ENDPOINT, files=files, data=data)
            image = None

            if image_json is None:
                worst_status = {STATUS: "image not uploaded " + str(record)}
                continue
            elif "detail" in image_json and "already exists" in image_json["detail"]:
                image, worst_status = self.parse_already_inserted(image_json["detail"])
            elif isinstance(image_json, list) and "already exists" in image_json[0]:
                image, worst_status = self.parse_already_inserted(image_json[0])
            elif ID not in image_json:
                worst_status = {STATUS: "image not uploaded " + str(record)}
                continue
            else:
                image = Image(self.token, self.endpoint, image_json)

            if REAL_IMAGE in record:
                image.set_real(record[REAL_IMAGE])

            if LABELS in record:
                for label_id in record[LABELS]:
                    image.add_label(label_id)

            if VALUES in record:
                for label_id in record[VALUES].keys():
                    image.add_label(label_id, record[VALUES][label_id])

            images.append(image)
        return images, worst_status

    def _create_image_data(self, record, noresize, noresize_on_server, test_image, metadata):
        if IMG_DATA in record:
            return {
                "base64": self.cv2img_to_base64(record[IMG_DATA], record[COLOR_SPACE], resize=not noresize),
                NORESIZE: noresize_on_server,
                TEST_IMAGE: test_image,
                META_DATA: metadata,
            }

        if FILE in record:
            # We cannot send files to request along with json data (for workspace)
            # That is why we load image from disk to base64 representation
            return {
                "base64": self.load_base64_file(record[FILE], resize=not noresize),
                NORESIZE: noresize_on_server,
                TEST_IMAGE: test_image,
                META_DATA: metadata,
            }

        if BASE64 in record:
            return {
                "base64": record[BASE64].decode("utf-8"),
                NORESIZE: noresize_on_server,
                TEST_IMAGE: test_image,
                META_DATA: metadata,
            }

        if URL in record:
            return {
                "base64": self.load_url_image(record[URL], resize=not noresize),
                NORESIZE: noresize_on_server,
                TEST_IMAGE: test_image,
                META_DATA: metadata,
            }

        raise Exception("No _file, _url, _base64, _img_data in record!")

    def construct_data(self, records=[], task_id=None, version=None, store_images=None):
        if len(records) == 0:
            raise Exception("Please specify at least one record in classify method!")

        if task_id is None:
            raise Exception("Please specify task")

        data = {RECORDS: self.preprocess_records(records), TASK_ID: task_id, VERSION: version}

        if store_images:
            data[STORE_IMAGES] = True

        return data

    def classify_on_task(self, records=[], task_id=None, version=None, store_images=None):
        """
        Takes the images and calls the ximilar client for classifying these images on the task.

        Usage:
            client = RecognitionClient('__YOUR_API_TOKEN__')
            result = client.classify_on_task({'_url':'__SOME_IMG_URL__'}, task_id="__UUID__")

        :param records: array of json/dicts [{'_url':'url-path'}, {'_file': ''}, {'_base64': 'base64encodeimg'}]
        :param task_id: id of task
        :param version: optional(integer of specific version), default None/production_version
        :param store_images: if true then store the images on the backend (available in higher plans)
        :return: json response
        """
        # version is default set to None, so ximilar will determine which one to take
        data = self.construct_data(records=records, task_id=task_id, version=version, store_images=store_images)
        result = self.post(self.PREDICT_ENDPOINT, data=data)
        try:
            self.check_json_status(result)
            return result
        except Exception as e:
            return None


class Task(RecognitionClient):
    """
    Task entity from /recognition/v2/task endpoint.
    Every task can have multiple recognition labels.
    """

    def __init__(self, token, endpoint, task_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = task_json[ID]
        self.name = task_json[NAME]
        self.type = task_json[TYPE]
        self.production_version = task_json[PRODUCTION_VERSION]
        self.workspace = task_json[WORKSPACE] if WORKSPACE in task_json else DEFAULT_WORKSPACE
        self.description = task_json[DESCRIPTION] if DESCRIPTION in task_json else ""
        self.last_train_status = task_json[LAST_TRAIN_STATUS] if LAST_TRAIN_STATUS in task_json else ""

    def __str__(self):
        return self.name

    def train(self):
        """
        Create new training/model and add it to the queue.
        :return: None
        """
        return self.post(TASK_ENDPOINT + self.id + "/train/")

    def remove(self):
        """
        Delete the recognition task from ximilar.
        """
        self.remove_task(self.id)

    def get_negative_label(self):
        """
        If the task is Tagging/Multi-Label then this will return the negative label of the Task.
        """
        labels, _ = self.get_labels()

        for label in labels:
            if label.negative_for_task:
                return label, RESULT_OK

        return None, RESULT_OK

    def get_labels(self):
        """
        Get labels of this task.
        :return: list of Labels
        """
        if LABELS in self.cache:
            return self.cache[LABELS], RESULT_OK
        else:
            labels, result = self.get_all_labels(suffix="?task=" + self.id)

            if result[STATUS] == STATUS_OK:
                self.cache[LABELS] = labels

            return self.cache[LABELS], result

    def get_label_by_name(self, name):
        """
        Get label with specified name which also belongs to this task.
        """
        labels, result = self.get_labels()
        if result[STATUS] == STATUS_OK:
            for label in labels:
                if label.name == name:
                    return label, RESULT_OK
        else:
            return None, result

        return None, {STATUS: "Label with this name not found!"}

    def classify(self, records, version=None, store_images=None):
        """
        Takes the images and calls the ximilar client for classifying these images on the task.

        Usage:
            client = RecognitionClient('__YOUR_API_TOKEN__')
            task, status = client.get_task('__ID_OF_TASK__')
            result = task.classify({'_url':'__SOME_IMG_URL__'})

        :param records: array of json/dicts [{'_url':'url-path'}, {'_file': ''}, {'_base64': 'base64encodeimg'}]
        :param version: optional(integer of specific version), default None/production_version
        :param store_images: if true then store the images on the backend (available in higher plans)
        :return: json response
        """
        # version is default set to None, so ximilar will determine which one to take
        data = self.construct_data(records=records, task_id=self.id, version=version, store_images=store_images)
        result = self.post(self.PREDICT_ENDPOINT, data=data)

        self.check_json_status(result)
        return result

    def add_label(self, label_id):
        """
        Add label to this task.
        :param label_id: identification of label
        :return: json/dict result
        """
        data = {LABEL_ID: label_id}
        return self.post(TASK_ENDPOINT + self.id + "/add-label/", data=data)

    def detach_label(self, label_id):
        """
        Remove/Detach label from the task.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id + "/remove-label/", data={LABEL_ID: label_id})

    def to_json(self):
        labels, status = self.get_labels()
        return {
            TASK_ID: self.id,
            NAME: self.name,
            TASK_TYPE: self.type,
            DESCRIPTION: self.description,
            LABELS: [label.id for label in labels],
        }


class Model(RecognitionClient):
    """
    Model entity from /recognition/v2/mode endpoint.
    Every model represents model/neural network which is trained or waiting for training.
    """

    def __init__(self, token, endpoint, model_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = model_json[ID]
        self.task_id = model_json[TASK]
        self.task_name = model_json[TASK_NAME]
        self.version = model_json[VERSION]
        self.train_status = model_json[TRAIN_STATUS]

    def remove(self):
        self.delete_model(self.id)

    def __str__(self):
        return self.task_name + " (" + self.train_status + ")"


class Label(RecognitionClient):
    """
    Label entity from /recognition/v2/label endpoint.
    Every label can be assigned to multiple tasks.
    """

    def __init__(self, token, endpoint, label_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = label_json[ID]
        self.name = label_json[NAME]
        self.type = label_json[TYPE] if TYPE in label_json else None
        self.tasks_count = label_json[TASKS_COUNT] if TASKS_COUNT in label_json else 0
        self.negative_for_task = label_json[NEGATIVE_FOR_TASK] if NEGATIVE_FOR_TASK in label_json else None
        self.workspace = label_json[WORKSPACE] if WORKSPACE in label_json else DEFAULT_WORKSPACE
        self.images_count = label_json[IMAGES_COUNT] if IMAGES_COUNT in label_json else -1
        self.description = label_json[DESCRIPTION] if DESCRIPTION in label_json else ""
        self.output_name = (
            label_json[OUTPUT_NAME] if OUTPUT_NAME in label_json and label_json[OUTPUT_NAME] else self.name
        )

    def __str__(self):
        return self.name

    def wipe(self):
        """
        Delete label and all images associated with this label.
        :return: None
        """
        self.delete(LABEL_ENDPOINT + self.id + "/wipe")

    def get_images_count(self):
        """
        Get count of the images connected to this label.
        :return:
        """
        if self.images_count >= 0:
            return self.images_count

        label_json = self.get(LABEL_ENDPOINT + self.id)
        self.images_count = label_json[IMAGES_COUNT]
        return self.images_count

    # TODO: get object count
    # TODO: get objects of given label

    def get_training_images(self, page_url=None, verification=None, not_label=False, test=False, value=None):
        """
        Get paginated result of images for specific label.

        :param page_url: optional, select the specific page of images, default first page
        :return: (list of images, next_page)
        """
        if page_url is None:
            if not_label:
                page_url = IMAGE_ENDPOINT + "?not_labels=" + self.id
            else:
                page_url = IMAGE_ENDPOINT + "?labels=" + self.id

            if value is not None:
                page_url += f"&label_value_{self.id}__gte={value}&label_value_{self.id}__lte={value}"
            print(page_url)
        return super().get_training_images(page_url=page_url, verification=verification, test=test)

    def upload_images(self, records):
        """
        Upload image to the ximilar.com and adding label to this image.
        :param records: list of files with _url, _base64, _file.
        :return: None
        """
        for i in range(len(records)):
            if LABELS in records[i]:
                records[i][LABELS].append(self.id)
            else:
                records[i][LABELS] = [self.id]
        return super().upload_images(records)

    def detach_image(self, image_id):
        """
        Remove/Detach label from the image.
        :param image_id: id of label
        :return: result
        """
        return self.post(IMAGE_ENDPOINT + image_id + "/remove-label/", data={LABEL_ID: self.id})

    def remove(self):
        return self.remove_label(self.id)

    def add_annotate_task(self, task_id):
        """
        Add task to this label.
        ! this is only for ANNOTATE (parent => child mapping behaviour)

        :param task_id: identification of label
        :return: json/dict result
        """
        return self.post(LABEL_ENDPOINT + self.id + "/add-task/", data={TASK_ID: task_id})

    def detach_annotate_task(self, task_id):
        """
        Remove/Detach task from the label.
        ! this is only for ANNOTATE (parent => child mapping behaviour)

        :param task_id: identification of label
        :return: json/dict result
        """
        return self.post(LABEL_ENDPOINT + self.id + "/remove-task/", data={TASK_ID: task_id})

    def get_annotate_tasks(self):
        data = self.get(LABEL_ENDPOINT + self.id)
        return [self.get_task(task["id"])[0] for task in data[RECOGNITION_TASKS]], STATUS_OK

    def to_json(self):
        return {
            LABEL_ID: self.id,
            NAME: self.name,
            NEGATIVE_FOR_TASK: self.negative_for_task,
            DESCRIPTION: self.description,
            OUTPUT_NAME: self.output_name,
            TYPE: self.type,
        }


class Image(RecognitionClient):
    """
    Image entity from /recognition/v2/image endpoint.
    Every image can have multiple recognition labels.
    """

    def __init__(self, token, endpoint, image_json):
        super().__init__(token, endpoint, resource_name=None)

        self.id = image_json[ID]
        self.created = image_json[CREATED]
        self.img_path = image_json[IMG_PATH]
        self.thumb_img_path = image_json[THUMB_IMG_PATH]
        self.verifyCount = image_json[VERIFY_COUNT] if VERIFY_COUNT in image_json else -1
        self.workspace = image_json[WORKSPACE] if WORKSPACE in image_json else DEFAULT_WORKSPACE
        self.img_width = image_json[IMG_WIDTH] if IMG_WIDTH in image_json else None
        self.img_height = image_json[IMG_HEIGHT] if IMG_HEIGHT in image_json else None
        # if the meta data was downloaded from the server, this field is never None
        self.meta_data = (
            None if META_DATA not in image_json else (image_json[META_DATA] if image_json[META_DATA] else {})
        )
        # file path after calling download_image on this object
        self._file = None
        self._objects = []
        self.test_image = image_json[TEST_IMAGE] if TEST_IMAGE in image_json else None
        self.real_image = image_json[REAL_IMAGE] if REAL_IMAGE in image_json else None

    def __str__(self):
        return self.thumb_img_path

    def remove(self):
        """
        Remove image from system.
        """
        return self.remove_image(self.id)

    def set_test(self, test):
        result = self.modify_images([self.id], test=test)
        self.test_image = test
        return result

    def set_real(self, real):
        self.modify_images([self.id], real=real)
        self.real_image = real

    def get_labels(self):
        """
        Get labels assigned to this image.
        :return: list of Labels
        """
        if LABELS not in self.cache:
            labels = [Label(self.token, self.endpoint, label) for label in self.get(IMAGE_ENDPOINT + self.id)[LABELS]]
        else:
            labels = self.cache[LABELS]

        return labels, RESULT_OK

    def add_label(self, label_id, value=None):
        """
        Add label to the image.
        :param label_id: id of label
        :return: result
        """
        data = {LABEL_ID: label_id} if value is None else {LABEL_ID: label_id, "value": value}
        return self.post(IMAGE_ENDPOINT + self.id + "/add-label/", data=data)

    def update_label(self, label_id, value=None):
        """
        Update value of the label on image.
        """
        data = {LABEL_ID: label_id, "value": value}
        return self.post(IMAGE_ENDPOINT + self.id + "/update-label/", data=data)

    def detach_label(self, label_id):
        """
        Remove/Detach label from the image.
        :param label_id: id of label
        :return: result
        """
        return self.post(IMAGE_ENDPOINT + self.id + "/remove-label/", data={LABEL_ID: label_id})

    def remove_label(self, label_id):
        return self.detach_label(label_id)

    def download(self, destination="", name=None):
        """
        Download image to the destination and store path to the _file of the object.
        :param destination: path on the disk
        :return: path of the file locally
        """
        self._file = super().download_image(self.img_path, destination=destination, name=name)
        return self._file

    def _ensure_meta_data(self):
        """
        If the meta_data were not downloaded yet, do so
        """
        if not self.meta_data:
            image_data = self.get(IMAGE_ENDPOINT + self.id)
            if META_DATA in image_data:
                self.meta_data = image_data[META_DATA]
        if not self.meta_data:
            self.meta_data = {}

    def get_meta_data(self):
        """
        Return the image meta data (dictionary) or empty dictionary
        :return: None is never returned
        """
        self._ensure_meta_data()
        return self.meta_data

    def add_meta_data(self, meta_data):
        """
        Add some meta data to image (extends already present meta data).
        """
        self._ensure_meta_data()
        if meta_data is None or not isinstance(meta_data, dict):
            raise Exception("Please specify dictionary of meta_data as param!")

        new_data = dict(list(self.meta_data.items()) + list(meta_data.items()))
        result = self.put(IMAGE_ENDPOINT + self.id, data={META_DATA: new_data})
        self.meta_data = result[META_DATA]
        return True

    def clear_meta_data(self):
        """
        Clear all meta data of image.
        """
        result = self.put(IMAGE_ENDPOINT + self.id, data={META_DATA: {}})
        self.meta_data = result[META_DATA]
        return True

    def extract_object_data(self, object_bbox, image_download_dir=""):
        """
        Extracting object/bounding box data from image.
        :param object_bbox: [xmin, ymin, xmax, ymax]
        :param image_download_dir directory to download the image before cutting the object
        :return: dict / { "img_data": [[]...], "color_space": "RGB"}
        """
        assert len(object_bbox) == 4

        self.download(image_download_dir)
        image = self.cv2_imread(self._file)
        return {
            IMG_DATA: image[int(object_bbox[1]) : int(object_bbox[3]), int(object_bbox[0]) : int(object_bbox[2])],
            COLOR_SPACE: "BGR",
        }

    def get_verifications(self):
        json_results = self.get("annotate/v2/verification/?image=" + self.id)
        self.verifyCount = len(json_results["results"])
        return json_results["results"]

    def verify(self, user_id):
        """
        Verify this image by some user.
        """
        result = self.post("annotate/v2/verification/", data={USER: user_id, IMAGE_ID: self.id})
        self.get_verifications()
        return result

    def unverify(self):
        """
        Unverify all users from image.
        """
        results = self.get_verifications()
        for result in results:
            self.delete("annotate/v2/verification/" + str(result["id"]))
        self.verifyCount = 0
        return True

    def to_json(self):
        labels, status = self.get_labels()
        return {
            IMAGE: self.id,
            LABELS: [label.id for label in labels],
            LABEL_NAMES: [label.output_name for label in labels],
            META_DATA: self.meta_data,
            IMG_HEIGHT: self.img_height,
            IMG_WIDTH: self.img_width,
            VERIFY_COUNT: self.verifyCount,
            FILE: self._file,
            IMG_PATH: self.img_path,
        }


class Workspace(RecognitionClient):
    """
    Workspace entity. All Task, Labels and Images are mapped to some workspace.
    Every workspace has some owner.
    """

    def __init__(self, token, endpoint, workspace_json):
        super().__init__(token, endpoint, workspace_json[ID])
        self.id = workspace_json[ID]
        self.name = workspace_json[NAME]
        self.owner = workspace_json[OWNER] if OWNER in workspace_json else None

    def __str__(self):
        return "Worskpace: (%s) (%s)" % (self.name, self.id)


class Point(RecognitionClient):
    """
    Workspace entity. All Task, Labels and Images are mapped to some workspace.
    Every workspace has some owner.
    """

    def __init__(self, token, endpoint, point_json):
        super().__init__(token, endpoint)
        self.id = point_json[ID]
        self.data = point_json[DATA]
        self.name = point_json[NAME]

    def __str__(self):
        return "POINT: (%s) (%s)" % (self.name, self.id)
